# -*- coding: utf-8 -*-
# Editobj3
# Copyright (C) 2007-2014 Jean-Baptiste LAMY

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
editobj3.introsp -- Introspection framework
-------------------------------------------


"""

import os, os.path, types, inspect, locale
import editobj3, editobj3.undoredo as undoredo
from collections import defaultdict

_entity_classes = ()
def _init_for_owlready2():
  global _entity_classes
  import owlready2
  _entity_classes = (owlready2.ThingClass, owlready2.Thing, owlready2.PropertyClass)
  import editobj3.observe
  editobj3.observe._has_owlready2 = True
  
PROPERTY_TYPE_NAMES      = {"property", "getset_descriptor"}
IGNORED_ATTRS            = {"__weakref__", "__abstractmethods__"}
BOOL_ATTRS               = {"visible", "hidden", "active"}
DEFAULT_ADD_METHODS      = ["insert", "add", "append"]
DEFAULT_REMOVE_METHODS   = ["remove", "discard"]
DEFAULT_HAS_ITEM_METHODS = ["has"]
_NEXT_PRIORITY           = 0

def _get_methods_for_attr(base_methods, attr):
  if attr == "__list__": return base_methods
  methods = ["%s_%s" % (method, attr) for method in base_methods]
  if attr.endswith("s"): methods = [method[:-1] for method in methods] + methods
  return methods

def _method_has_nb_args(method, nb):
  try: return method.__code__.co_argcount == nb
  except: return False
  

def print_info(o):
  """Prints the information available from introsp for the given object.

:param o: the object."""
  
  descr = description_for_object(o)
  print("Introsp info about:", o)
  print("Label        :", descr.label_for(o))
  print("Details      :", descr.details_for(o))
  print("Icon filename:", descr.icon_filename_for(o))
  print("Actions      :", ", ".join(action.label for action in descr.actions_for(None, None, o)))
  property_attributes = descr.property_attributes_of(o)
  if property_attributes:
    print("Property attributes:", ", ".join(attribute.name for attribute in property_attributes))
  print("Attributes:")
  for attribute in descr.attributes_of(o):
    field_class = attribute.field_class_for(o)
    unit        = attribute.unit_for(o)
    print("  %s: %s" % (attribute.name, field_class))
    if unit: print("  %s: %s" % (attribute.name, unit))
    if field_class.multiple_object_field:
      print("  %s: addable values: %s" % (attribute.name, attribute.addable_values_to(o)))
      print("  %s: can add to: %s, can remove from: %s, can reorder: %s" % (attribute.name, attribute.can_add_to(o), attribute.can_remove_from(o), attribute.can_reorder(o)))
      print("  %s: actions on a child:" % attribute.name, ", ".join(action.label for action in descr.actions_for(o, attribute, None)))
      

class Attribute(object):
  def __init__(self, name, field_class = "auto", unit = "", priority = 0, inverse_attr = None, optional = True, get_method = "auto", set_method = "auto", addable_values = None, add_method = "auto", remove_method = "auto", reorder_method = "auto", has_item_method = "auto", label = None):
    self.name            = name
    self.field_class     = field_class
    self.unit            = unit
    self.priority        = priority
    self.inverse_attr    = inverse_attr
    self.optional        = optional
    self.get_method      = get_method
    self.set_method      = set_method
    self.addable_values  = addable_values
    self.add_method      = add_method
    self.remove_method   = remove_method
    self.reorder_method  = reorder_method
    self.has_item_method = has_item_method
    self.allow_none      = False
    
    if (self.get_method and (self.get_method != "auto")) or (self.set_method and (self.set_method != "auto")):
      self.is_property = True
      self.optional    = False # Properties cannot be optional
    else:
      self.is_property = False
      
    if label is None: label = name.replace("_", " ")
    if label: self.label = editobj3.TRANSLATOR(label)
    else:     self.label = ""
    
  def update(self, **kargs):
    if ("label" in kargs) and kargs["label"]: kargs["label"] = editobj3.TRANSLATOR(kargs["label"])
    self.__dict__.update(kargs)
    
    if (self.get_method and (self.get_method != "auto")) or (self.set_method and (self.set_method != "auto")):
      self.is_property = True
      self.optional    = False # Properties cannot be optional
    else:
      self.is_property = False
      
  def __eq__  (self, other): return (self.name == other.name) and (self.__class__ is other.__class__)
  def __hash__(self): return hash((self.name, self.__class__))
    
  def copy(self):
    return Attribute(self.name, self.field_class, self.unit, self.priority, self.inverse_attr, self.optional, self.get_method, self.set_method, self.addable_values, self.add_method, self.remove_method, self.reorder_method, self.has_item_method, self.label)
    
  def field_class_for(self, o, dont_get_value = False):
    if self.field_class != "auto": return self.field_class
    
    # Default heuristic rules
    import editobj3.field as field
    if self.name == "__list__": return field.HierarchyOrObjectListField
    if self.name.startswith("_"): return field.HiddenField

    if dont_get_value and (self.has_item_method != "auto"):
      return field.HierarchyOrObjectSelectorField
    val = self.get_value_for(o)
    if   isinstance(val, bool ): return field.BoolField
    elif isinstance(val, float): return field.FloatField
    elif isinstance(val, int):
      if ((val == 1) or (val == 0)) and (self.name.startswith("is") or self.name.startswith("has") or self.name.endswith("enabled") or self.name.endswith("Enabled") or (self.name in BOOL_ATTRS)): return field.BoolField
      return field.IntField
    elif isinstance(val, str):
      if self.name == "password": return field.PasswordField
      if self.name.endswith("filename") or self.name.endswith("Filename"): return field.FilenameField
      if self.name.endswith("url") or self.name.endswith("Url") or self.name.endswith("URL"): return field.URLField
      if (len(val) > 200) or ("\n" in val): return field.TextField
      else:                                 return field.StringField
    elif isinstance(val, list) or isinstance(val, set) or isinstance(val, tuple) or isinstance(val, dict): return field.HierarchyOrObjectListField
    elif isinstance(val, object) and hasattr(val, "__dict__"):
      if (self.name == "parent") or (self.name == "master") or (self.name == "root"): return field.HiddenField
      return field.HierarchyOrObjectSelectorField
    return field.HiddenField
    
  def unit_for    (self, o): return self.unit
  def priority_for(self, o): return self.priority
  
  def _call_method(self, method, o, *args):
    if callable(method):     return method(o, *args)
    if method:               return getattr(o, method)(*args)
    raise AttributeError("No such method")
    
  def _call_insert_or_add_method(self, method, o, value, index):
    if index is None:
      if self.name == "__list__": index = len(o)
      else:                       index = len(self.get_value_for(o))
    if not callable(method): method = getattr(o.__class__, method)
    try:              return method(o, index, value)
    except TypeError: return method(o, value)
    
  def get_value_for(self, o, default = None):
    if self.get_method == "auto": return getattr(o, self.name, default)
    return self._call_method(self.get_method, o)
    
  def can_set_value_for(self, o): return bool(self.set_method)
  
  def set_value_for(self, o, value):
    if self.set_method == "auto": return setattr(o, self.name, value)
    return self._call_method(self.set_method, o, value)
    
  def addable_values_to(self, o):
    if not self.addable_values: return []
    if   callable  (self.addable_values):      addable_values = self.addable_values(o)
    elif isinstance(self.addable_values, str): addable_values = getattr(o, self.addable_values)()
    else:                                      addable_values = self.addable_values
    return sorted(addable_values, key = lambda value: (-int(str(value).startswith("(")), locale.strxfrm(str(value))))
    
  def can_add_to(self, o):
    if self.add_method != "auto": return bool(self.add_method)
    for method in _get_methods_for_attr(DEFAULT_ADD_METHODS, self.name):
      if hasattr(o, method): return True
    values = self.get_value_for(o, self.name)
    if (values is o) or (self.name == "__list__"): return False
    return description_for_object(values)._get_attribute("__list__").can_add_to(values)
    
  def add_to(self, o, value, index = None):
    if self.add_method != "auto": return self._call_insert_or_add_method(self.add_method, o, value, index)
    for method in _get_methods_for_attr(DEFAULT_ADD_METHODS, self.name):
      if hasattr(o, method): return self._call_insert_or_add_method(method, o, value, index)
    values = self.get_value_for(o, self.name)
    if (values is o) or (self.name == "__list__"): return False
    return description_for_object(values)._get_attribute("__list__").add_to(values, value, index)
    
  def can_remove_from(self, o):
    if self.remove_method != "auto": return bool(self.remove_method)
    for method in _get_methods_for_attr(DEFAULT_REMOVE_METHODS, self.name):
      if hasattr(o, method): return True
    values = self.get_value_for(o, self.name)
    if (values is o) or (self.name == "__list__"): return False
    return description_for_object(values)._get_attribute("__list__").can_remove_from(values)
    
  def remove_from(self, o, value):
    if self.remove_method != "auto": return self._call_method(self.remove_method, o, value)
    for method in _get_methods_for_attr(DEFAULT_REMOVE_METHODS, self.name):
      if hasattr(o, method): return getattr(o, method)(value)
    values = self.get_value_for(o, self.name)
    if (values is o) or (self.name == "__list__"): return False
    return description_for_object(values)._get_attribute("__list__").remove_from(values, value)
    
  def has_item_for(self, o):
    if self.has_item_method != "auto": return self._call_method(self.has_item_method, o)
    for method in _get_methods_for_attr(DEFAULT_HAS_ITEM_METHODS, self.name):
      if hasattr(o, method): return getattr(o, method)()
    return self.get_value_for(o)
    
  def can_reorder(self, o):
    if self.reorder_method != "auto": return bool(self.reorder_method)
    return isinstance(self.get_value_for(o), list)
    
  def reorder(self, o, value, direction):
    if self.reorder_method != "auto": return self._call_method(self.reorder_method, o, value, direction)
    values = self.get_value_for(o)
    index = values.index(value)
    if direction == -1:
      if index > 0:               values[index - 1], values[index] = values[index], values[index - 1]
    else:
      if index < len(values) - 1: values[index + 1], values[index] = values[index], values[index + 1]
      
  def __repr__(self):
    if   self.get_method == "auto":     get = ""
    elif self.get_method:               get = ", get with '%s'" % self.get_method
    else:                               get = ", write-only"
    if   self.set_method == "auto":     set = ""
    elif self.set_method:               set = ", set with '%s'" % self.set_method
    else:                               set = ", read-only"
    return "<Attribute '%s'%s%s>" % (self.name, get, set)
    


_TYPE_DESCRS  = {}
_CLASS_DESCRS = {}
def description_for_class(klass):
  """Returns the class description for the given class.
The class description is created if not already existing.

:param Klass: the class.
:returns: a class description (:class:`ClassDescr`).
"""
  descr = _CLASS_DESCRS.get(klass)
  if not descr: descr = _CLASS_DESCRS[klass] = ClassDescr(klass, getattr(klass, "__mro__", None) or klass.mro())
  return descr
description = description_for_class

def description_for_type(type):
  """Returns the description for the given type.
The description is created if not already existing.

:param type: the type.
:returns: a class description (:class:`ClassDescr`).
"""
  descr = _TYPE_DESCRS.get(type)
  if not descr: descr = _TYPE_DESCRS[type] = TypeDescr(type, getattr(type, "__mro__", None) or type.mro())
  return descr

def description_for_object(o):
  """Returns the class description for the given object.
The class description is created if not already existing.

:param o: the object.
:returns: a class description (:class:`ClassDescr`).
"""
  if isinstance(o, type):
    return description_for_type(o)
  else:
    return description_for_class(o.__class__)

class ClassDescr(object):
  """A class description. It describes the attributes of the class and their types, the associated icon,...
Editobj3 is able to automatically guess the attributes and their types, but you can overide the guessed
value by calling various methods of the class description.

.. warning::
   You should not create ClassDescr, but rather call :func:`description`."""
  def _get_parent_description(self, klass): return description_for_class(klass)
  
  def __init__(self, klass, inherited_classes):
    self.klass            = klass
    self.attributes       = {}
    self.actions          = {}
    self.cactions         = {}
    self.icon_filename    = None
    self.label            = None
    self.details          = None
    self.constructor      = None
    if not object is inherited_classes[-1]: inherited_classes = inherited_classes + [object]
    self.inherited_descrs = [self._get_parent_description(klass) for klass in inherited_classes[1:]]
    self.inherited_descrs.insert(0, self)
    
    self.attributes.update(self.search_attributes(klass))
    
    self.is_owl_entity = isinstance(klass, _entity_classes)
    
  def create_attribute(self, name):
    if name in self.attributes: return self.attributes[name]
    for descr in self.inherited_descrs:
      if name in descr.attributes:
        attribute = descr.attributes[name].copy()
        break
    else: attribute = Attribute(name)
    return attribute
    
  def search_attributes(self, klass):
    attributes = {}
    for attr in dir(klass):
      if attr in IGNORED_ATTRS: continue
      kval = getattr(klass, attr)
      
      if   type(kval).__name__ in PROPERTY_TYPE_NAMES:
        attribute             = attributes.get(attr) or self.create_attribute(attr)
        attribute.is_property = True
        attribute.optional    = False
        if isinstance(kval, property):
          if not kval.fset: attribute.set_method = None
          if not kval.fget: attribute.get_method = None
        else:
          if not kval.__set__: attribute.set_method = None
          if not kval.__get__: attribute.get_method = None
        attributes[attribute.name] = attribute
        
      elif attr.startswith("set") and (len(attr) > 3) and _method_has_nb_args(kval, 2):
        if attr[3] == "_": attr2 = attr[4:]
        else:              attr2 = attr[3].lower() + attr[4:]
        attribute             = attributes.get(attr2) or self.create_attribute(attr2)
        attribute.set_method  = attr
        attribute.is_property = True
        attributes[attribute.name] = attribute
        
      elif attr.startswith("get") and (len(attr) > 3) and _method_has_nb_args(kval, 1):
        if attr[3] == "_": attr2 = attr[4:]
        else:              attr2 = attr[3].lower() + attr[4:]
        attribute             = attributes.get(attr2) or self.create_attribute(attr2)
        attribute.get_method  = attr
        attribute.is_property = True
        attribute.optional    = False
        attributes[attribute.name] = attribute
    return attributes
  
  def __repr__(self): return ("<%s for %s, attributes:\n  " % (self.__class__.__name__, self.klass)) + "\n  ".join([repr(attribute) for attribute in self.attributes.values()]) + "\n>"
  
  def set_label(self, label):
    """Set the label for this class. If no label is provided, the str() function is used.

:param label: either a string (the label) or a callable (taking one argument, the object, and returning the label)."""
    self.label = label
    
  def set_details(self, details):
    """Set the details for this class. If no details are provided, an empty string is used.

:param details: either a string (the details) or a callable (taking one argument, the object, and returning the details)."""
    self.details = details
    
  def set_icon_filename(self, icon_filename):
    """Set the icon filename for this class. If no icon is provided, a Python icon is used.

:param icon_filename: either a string (the icon filename) or a callable (taking one argument, the object, and returning the icon filename)."""
    self.icon_filename = icon_filename

  def set_constructor(self, constructor):
    """Set the constructor for this class. By default, EditObj 3 creates new instances by calling the class with a single argument: the parent objet.
If your class's __init__ does not work like that, you can either set a Constructor, or create a FormConstructor subclass if you need to ask some information to the user prior to create the instance.

Example of Constructor if the class's __init__ does not expect any argument:

>>> descr.set_constructor(introsp.Constructor(lambda klass, parent: klass()))

Example of a FormConstructor asking the user for a "name" argument:

>>> class MyFormConstructor(introsp.FormConstructor):
>>>   def __init__(self):
>>>     self.name = "default name"
>>> descr.set_constructor(MyFormConstructor)

:param constructor: a Constructor object, a FormConstructor subclass, or None for using the class itself as a constructor."""
    self.constructor = constructor
    
  #def def_attr(self, attr, field_class = "auto", unit = "auto", priority = "auto", inverse_attr = "auto", optional = "auto", get_method = "auto", set_method = "auto", addable_values = "auto", add_method = "auto", remove_method = "auto", reorder_method = "auto", has_item_method = "auto", label = None, allow_none = False):
  def def_attr(self, attr, field_class = 0, **kargs):
    """def_attr(attr, field_class = "auto", unit = "auto", priority = "auto", inverse_attr = "auto", optional = "auto", get_method = "auto", set_method = "auto", addable_values = "auto", add_method = "auto", remove_method = "auto", reorder_method = "auto", has_item_method = "auto", label = None)

This method is similar to :func:`editobj3.introsp.def_attr`, but it is restricted to the class described by this class description (including subclasses)."""
    global _NEXT_PRIORITY
    attribute = self.attributes[attr] = self.create_attribute(attr)
    
    if field_class != 0: kargs["field_class"] = field_class # Allows to give the field_class argument by position
    
    if not "priority" in kargs:
      attribute.priority = _NEXT_PRIORITY
      _NEXT_PRIORITY += 1
    elif kargs["priority"] is None: del kargs["priority"]
    
    if ("label" in kargs) and (kargs["label"]): kargs["label"] = editobj3.TRANSLATOR(kargs["label"])
    
    if kargs.get("field_class", 0) is None: raise ValueError("None is not a valid Field class; use editobj3.field.HiddenField")
    
    attribute.update(**kargs)
    
    # if field_class     != "auto": attribute.field_class     = field_class
    # if unit            != "auto": attribute.unit            = unit
    # if not priority    is  None :
    #   if priority      != "auto": attribute.priority        = priority
    #   else:                       attribute.priority        = _NEXT_PRIORITY; _NEXT_PRIORITY += 1
    # if inverse_attr    != "auto": attribute.inverse_attr    = inverse_attr
    # if optional        != "auto": attribute.optional        = optional
    # if get_method      != "auto": attribute.get_method      = get_method
    # if set_method      != "auto": attribute.set_method      = set_method
    # if addable_values  != "auto": attribute.addable_values  = addable_values
    # if add_method      != "auto": attribute.add_method      = add_method
    # if remove_method   != "auto": attribute.remove_method   = remove_method
    # if reorder_method  != "auto": attribute.reorder_method  = reorder_method
    # if has_item_method != "auto": attribute.has_item_method = has_item_method
    # if label           !=  None : attribute.label           = editobj3.TRANSLATOR(label)
    # return attribute
    
  def add_action(self, action):
    if isinstance(action, ActionOnAChild):
      if action.attr in self.cactions: self.cactions[action.attr][action.name] = action
      else:                            self.cactions[action.attr] = { action.name : action }
    else:                              self.actions [action.name] = action
    
  def create_new(self, parent):
    for descr in self.inherited_descrs:
      if descr.constructor: return descr.constructor.create_new(self.klass, parent)
    return self.klass(parent)
    
  # Per-instance methods
  
  def label_for(self, o):
    for descr in self.inherited_descrs:
      if descr.label:
        if callable(descr.label): return descr.label(o)
        else:                     return descr.label
    return str(o)
    
  def details_for(self, o):
    for descr in self.inherited_descrs:
      if descr.details:
        if callable(descr.details): return descr.details(o)
        else:                       return descr.details
    if hasattr(o, "details") or ("details" in self.attributes): return self._get_attribute("details").get_value_for(o)
    return ""
    
  def icon_filename_for(self, o):
    for descr in self.inherited_descrs:
      if descr.icon_filename:
        if callable(descr.icon_filename): return descr.icon_filename(o)
        else:                             return descr.icon_filename
    if hasattr(o, "icon_filename") or ("icon_filename" in self.attributes): return self._get_attribute("icon_filename").get_value_for(o)
    return os.path.join(editobj3._ICON_DIR, "python.svg")

  def attributes_of(self, o):
    attributes = { attribute.name : attribute # Keep only one attribute per name
                   for descr in reversed(self.inherited_descrs) 
                   for attribute in descr.attributes.values()
                   if attribute.optional == False }
    if hasattr(o.__class__, "__attrs__"):
      for attr in o.__attrs__():
        if not attr in attributes: attributes[attr] = self._get_attribute(attr, False)
    else:
      o_dict = getattr(o, "__dict__", None)
      if o_dict:
        for attr in o_dict:
          if not attr in attributes: attributes[attr] = self._get_attribute(attr, False)
    return sorted(attributes.values(), key = lambda attribute: (attribute.priority, locale.strxfrm(attribute.label)))
    
  def property_attributes_of(self, o = None):
    return set({ attribute.name : attribute for descr in reversed(self.inherited_descrs) for attribute in descr.attributes.values() if attribute.is_property }.values())
    
  def _get_attribute(self, attr, register = True):
    for descr in self.inherited_descrs:
      if attr in descr.attributes: return descr.attributes[attr]
    attribute = Attribute(attr)
    if register: self.attributes[attr] = attribute
    return attribute
    
  def actions_for(self, parent, attribute, o):
    actions = {}
    for descr in reversed(self.inherited_descrs): actions.update(descr.actions)
    actions = { action for action in actions.values() if action.filter(o) }
    
    for o_attribute in self.attributes_of(o):
      field_class = o_attribute.field_class_for(o)
      if field_class.object_field:
        if field_class.multiple_object_field:
          if(o_attribute.can_add_to       (o) and
             o_attribute.addable_values_to(o)): actions.add(AppendAction(o_attribute))
        elif(o_attribute.can_set_value_for(o) and
             o_attribute.addable_values_to(o)): actions.add(SetAction(o_attribute))
        
    if parent:
      cactions     = {}
      parent_descr = description_for_object(parent)
      for descr in reversed(parent_descr.inherited_descrs):
        if attribute.name in descr.cactions: cactions.update(descr.cactions[attribute.name])
      actions.update(action for action in cactions.values() if action.filter(parent, attribute, o))
      
      field_class = attribute.field_class_for(parent)
      if field_class.object_field:
        if field_class.multiple_object_field:
          if(attribute.can_add_to       (parent) and
             attribute.addable_values_to(parent)): actions.add(InsertAction(attribute))
          if attribute.can_remove_from  (parent) : actions.add(_REMOVE_ACTION)
          if attribute.can_reorder      (parent) : actions.add(_MOVE_UP_ACTION); actions.add(_MOVE_DOWN_ACTION)
        else:
          if(attribute.can_set_value_for(parent) and
             attribute.addable_values_to(parent)): actions.add(ReplaceAction(attribute))
        
    actions = sorted(actions, key = lambda action: (action.priority, locale.strxfrm(action.label)))
    return actions
    
  def do_action(self, action, undo_stack, editor, o, *args): return action.do(undo_stack, editor, o, *args)

class TypeDescr(ClassDescr):
  def _get_parent_description(self, type): return description_for_type(type)
  

class Action(object):
  def __init__(self, name, func, filter = None, accept_object_pack = False, default = False, field_class = None, priority = 0):
    self.name                = name
    self.func                = func
    self.field_class         = field_class
    self.accept_object_pack  = accept_object_pack
    self.default             = default
    self.label               = editobj3.TRANSLATOR(self.name)
    self.priority            = priority
    if filter: self.filter   = filter
    
  def filter(self, o): return True
  
  def do(self, undo_stack, editor, o, *args):
    if callable(self.func): return self.func(undo_stack, editor, o, *args)
    else:                   return getattr(o, self.func)(undo_stack, editor, *args)
      
class ActionOnAChild(Action):
  def __init__(self, attr, name, func, filter = None, accept_object_pack = False, default = False, field_class = None, priority = 0):
    Action.__init__(self, name, func, filter, accept_object_pack, default, field_class, priority)
    self.attr = attr
    
  def filter(self, o, attribute, child): return True


class AddAction(object):
  def choose_new_child(self, editor, o, callback0, callback, label):
    addable_values = NamedList(self.attribute.addable_values_to(o), label)
    
    return_value = [None]
    def on_validate(new_child):
      if (new_child is None) or (new_child is addable_values): return
      if   isinstance(new_child, NewInstanceOf):
        new_child = new_child.create_new(o)
        if new_child is None: return
      elif isinstance(new_child, ObjectPack):
        for i in range(len(new_child.objects)):
          if isinstance(new_child.objects[i], NewInstanceOf): new_child.objects[i] = new_child.objects[i].create_new(o)
      callback0(new_child)
      callback (new_child)
      
    if   len(addable_values.children) >  1: editobj3.edit(addable_values, on_validate = on_validate, flat_list = True, width = 1024, height = 752, master = editor)
    elif len(addable_values.children) == 1: on_validate(addable_values.children[0])
    
  def do_insert(self, undo_stack, parent, index, new_child):
    if isinstance(new_child, ObjectPack): new_childs =  new_child.objects
    else:                                 new_childs = [new_child]
    
    def do_it():
      values = self.attribute.get_value_for(parent)
      for new_child in new_childs:
        if not new_child in values: # Else, already added by the constructor
          self.attribute.add_to(parent, new_child, index)
          
    def undo_it():
      for new_child in new_childs: self.attribute.remove_from(parent, new_child)
      
    undoredo.UndoableOperation(do_it, undo_it, editobj3.TRANSLATOR("add '%s'") % ", ".join(description_for_object(i).label_for(i) for i in new_childs), undo_stack)
    return new_child
    
  def do_set(self, undo_stack, o, new_value):
    old_value = self.attribute.get_value_for(o)
    
    def do_it  (): self.attribute.set_value_for(o, new_value)
    def undo_it(): self.attribute.set_value_for(o, old_value)
    
    undoredo.UndoableOperation(do_it, undo_it, editobj3.TRANSLATOR("set '%s'") % description_for_object(new_value).label_for(new_value), undo_stack)
    return new_value
  

class SetAction(Action, AddAction):
  def __init__(self, attribute):
    self.attribute = attribute
    Action.__init__(self, editobj3.TRANSLATOR("Set %s...") % attribute.label, None)
    
  def do(self, undo_stack, editor, o, callback):
    def callback0(new_child): self.do_set(undo_stack, o, new_child)
    return self.choose_new_child(editor, o, callback0, callback, editobj3.TRANSLATOR("Set %s to...") % self.attribute.label)
    
class AppendAction(Action, AddAction):
  def __init__(self, attribute):
    self.attribute = attribute
    Action.__init__(self, editobj3.TRANSLATOR("Add %s...") % attribute.label, None)
    
  def do(self, undo_stack, editor, o, callback):
    def callback0(new_child): self.do_insert(undo_stack, o, None, new_child)
    return self.choose_new_child(editor, o, callback0, callback, editobj3.TRANSLATOR("Add %s...") % self.attribute.label)
    
class ReplaceAction(ActionOnAChild, AddAction):
  def __init__(self, attribute):
    self.attribute = attribute
    ActionOnAChild.__init__(self, attribute.name, editobj3.TRANSLATOR("Replace %s...") % attribute.label, None)
    
  def do(self, undo_stack, editor, parent, attribute, o, callback):
    def callback0(new_child): self.do_set(undo_stack, parent, new_child)
    return self.choose_new_child(editor, parent, callback0, callback, editobj3.TRANSLATOR("Replace %s by...") % self.attribute.label)
    
class InsertAction(ActionOnAChild, AddAction):
  def __init__(self, attribute):
    self.attribute = attribute
    ActionOnAChild.__init__(self, attribute.name, editobj3.TRANSLATOR("Insert %s...") % attribute.label, None)
    
  def do(self, undo_stack, editor, parent, attribute, o, callback):
    def callback0(new_child):
      values = self.attribute.get_value_for(parent)
      if hasattr(values, "index"): index = values.index(o) + 1
      else:                        index = None
      self.do_insert(undo_stack, parent, index, new_child)
    return self.choose_new_child(editor, parent, callback0, callback, editobj3.TRANSLATOR("Insert %s...") % self.attribute.label)
    
class RemoveAction(ActionOnAChild):
  def __init__(self): ActionOnAChild.__init__(self, "__remove__", editobj3.TRANSLATOR("Remove"), None, accept_object_pack = True)
  
  def do(self, undo_stack, editor, parent, attributes, o):
    if isinstance(parent, ObjectPack): parents = parent.objects; objects = o.objects
    else:                              parents = [parent];       objects = [o];       attributes = [attributes]
    indexes    = [None] * len(parents)
    for i in range(len(parents)):
      values        = attributes[i].get_value_for(parents[i])
      if hasattr(values, "index"): indexes[i] = values.index(objects[i])
      
    def do_it():
      for i in range(len(parents)): attributes[i].remove_from(parents[i], objects[i])
      
    def undo_it():
      for i in range(len(parents)): attributes[i].add_to(parents[i], objects[i], indexes[i])
      
    labels = "', '".join([description_for_object(o).label_for(o) for o in objects])
    undoredo.UndoableOperation(do_it, undo_it, editobj3.TRANSLATOR("remove '%s'") % labels, undo_stack)

    
class MoveAction(ActionOnAChild):
  def __init__(self, name, direction):
    ActionOnAChild.__init__(self, "__move__", editobj3.TRANSLATOR(name), None, accept_object_pack = True)
    self.direction = direction
    
  def do(self, undo_stack, editor, parent, attributes, o):
    print(attributes)
    if isinstance(parent, ObjectPack): parents = parent.objects; objects = o.objects
    else:                              parents = [parent];       objects = [o];       attributes = [attributes]
    
    parents_attributes_objects = []
    for i in range(len(parents)):
      values    = attributes[i].get_value_for(parents[i])
      index     = values.index(objects[i])
      if (self.direction == -1) and (index <=               0): return # Cannot move up
      if (self.direction ==  1) and (index >= len(values) - 1): return # Cannot move down
      parents_attributes_objects.append((index, parents[i], attributes[i], objects[i]))
    parents_attributes_objects.sort(key = lambda index_parent_attribute_o: index_parent_attribute_o[0])
    
    def move_up():
      for index, parent, attribute, o in parents_attributes_objects: attribute.reorder(parent, o, -1)
      
    def move_down():
      for index, parent, attribute, o in reversed(parents_attributes_objects): attribute.reorder(parent, o, 1)
      
    labels = "', '".join([description_for_object(o).label_for(o) for o in objects])
    if self.direction == -1: undoredo.UndoableOperation(move_up  , move_down, editobj3.TRANSLATOR("move up '%s'"  ) % labels, undo_stack)
    else:                    undoredo.UndoableOperation(move_down, move_up  , editobj3.TRANSLATOR("move down '%s'") % labels, undo_stack)

    
_REMOVE_ACTION    = RemoveAction()
_MOVE_UP_ACTION   = MoveAction("Move up"  , -1)
_MOVE_DOWN_ACTION = MoveAction("Move down",  1)


class NewInstanceOf(object):
  def __init__(self, klass, label = None, icon_filename = None):
    self._klass = klass
    if label:         self._label = label
    else:             self._label = editobj3.TRANSLATOR("(new %s)" % klass.__name__)
    if icon_filename: self.icon_filename = icon_filename
  def __str__ (self): return self._label
  def __repr__(self): return "<NewInstanceOf for %s>" % self._klass.__name__
  def create_new(self, parent): return description(self._klass).create_new(parent)

class Constructor(object):
  def __init__(self, func):
    self._func  = func
  def create_new(self, klass, parent):
    return self._func(klass, parent)
  
class FormConstructor(object):
  @classmethod
  def create_new(FormClass, klass, parent):
    form = FormClass()
    form._klass = klass
    ok = [0]
    def on_validate(o):
      if o: ok[0] = 1
    editobj3.edit(form, on_validate = on_validate)
    if ok[0]: return form.create_new_from_form(klass, parent)
    
  def __str__ (self): return editobj3.TRANSLATOR("(new %s)" % self._klass.__name__)
  def __repr__(self): return "<FormConstructor for %s>" % self._klass.__name__
  
  def create_new_from_form(self, klass, parent):
    d = self.__dict__.copy()
    del d["_klass"]
    return klass(**d)


class NamedList(object):
  def __init__(self, children, name):
    self.children = children
    self._name    = name
  def __str__(self): return self._name
  
description(NamedList).def_attr("children", label = "", priority = None, add_method = None, remove_method = None, reorder_method = None)

add_action = description(object).add_action

def _edit_in_new_window_action(undo_stack, editor, o):
  import editobj3.editor
  editobj3.edit(o, undo_stack = undo_stack, master = editor)

_EDIT_ACTION = Action("Edit in new window...", _edit_in_new_window_action)
add_action(_EDIT_ACTION)


def def_attr(attr, field_class = 0, **kargs):
  """def_attr(attr, field_class = "auto", unit = "auto", priority = "auto", inverse_attr = "auto", optional = "auto", get_method = "auto", set_method = "auto", addable_values = "auto", add_method = "auto", remove_method = "auto", reorder_method = "auto", has_item_method = "auto", label = None)

Defines an attribute.

:param attr: the name of the attribute, or "__list__".
:param field_class: the class of field used to edit the attribute (see editobj3.field), or None to hide the attribute.
:param unit: the unit displayed on the right of the editing field (default to no unit).
:param priority: the priority value, used for ordering the attributes in the attribute pane (default: display the attributes
                 in the order they have been def_attr'ed); use None if you don't want to change the priority.
:param inverse_attr: the name of the inverse attribute (e.g. 'parent' attribute is tipically the inverse of 'children');
                     inverse attributes are automatically hidden in the tree view (default: not inverse attribute).
:param optional: the attribute may not be defined for all instances of the class
                 (default: True, except for property and get/set attributes).
:param get_method: the method for getting the value of the attribute.
:param set_method: the method for setting the value of the attribute.
:param addable_values: the list of possible values to add (to a list or set attribute),
                       or the value that the attribute can take for (for non-list object attribute).
                       Can also be a callable that returns the list of addable values for the given object (given in argument).
                       Use :class:`NewInstanceOf` for allowing the addition of fresh new instances of a given class.
:param add_method: the method for adding items to the attribute's value (only meaningful for list or set attribute)
                   Notice that ADD_METHOD also accept insert-like method (this should be preferred for ordered values).
:param remove_remove: the method for removing items from the attribute's value.
:param reorder_method: the method for reordering items in the attribute's value.
:param has_item_method: the method for testing the presence of items in the attribute's value.
:param label: the label for displaying the attribute (default to the attribute name).

All method parameters can be a method name, a callable, or None (to disable the corresponding functionality).

Except for the two first parameters, all parameters to def_attr() must be given by name, for example:

>>> introsp.def_attr("hit_points", field.IntField, unit = "HP")
>>> introsp.def_attr("items", field.ObjectListField,
                     addable_values = [NewInstanceOf(Item)],
                     add_method = "add_item",
                     )
"""
  found = False
  for descr in _CLASS_DESCRS.values():
    if attr in descr.attributes:
      descr.def_attr(attr, field_class, **kargs)
      found = True
  if not found: description(object).def_attr(attr, field_class, **kargs)

description(list ).def_attr("__list__", label = "", priority = None, optional = False, get_method = lambda o: o, set_method = None)
description(tuple).def_attr("__list__", label = "", priority = None, optional = False, get_method = lambda o: o, set_method = None, reorder_method = None, add_method = None, remove_method = None)
description(set  ).def_attr("__list__", label = "", priority = None, optional = False, get_method = lambda o: o, set_method = None, reorder_method = None)


class ObjectPack(object):
  def __init__(self, objects):
    self.objects             = objects
    self.attributes          = [description_for_object(o).attributes_of         (o) for o in objects]
    self.property_attributes = [description_for_object(o).property_attributes_of(o) for o in objects]
    self.attrs               = [{ attribute.name for attribute in o_attributes } for o_attributes in self.attributes]
    self.property_attrs      = [{ attribute.name for attribute in o_attributes } for o_attributes in self.property_attributes]
    
class NonConsistent(object):
  def __repr__(self): return "NonConsistent"
  def __bool__(self): return False
NonConsistent = NonConsistent()

class Inexistent(object):
  def __repr__(self): return "Inexistent"
  def __bool__(self): return False
Inexistent = Inexistent()

_OBJECT_PACK_ATTRIBUTES = {}
def _object_pack_attribute(name): return _OBJECT_PACK_ATTRIBUTES.get(name) or ObjectPackAttribute(name)

class ObjectPackAttribute(Attribute):
  def __init__(self, name):
    Attribute.__init__(self, name)
    _OBJECT_PACK_ATTRIBUTES[name] = self
    
  def unit_for(self, pack):
    unit = ""
    for i in range(len(pack.objects)):
      if self.name in pack.attrs[i]:
        new_unit = description_for_object(pack.objects[i])._get_attribute(self.name).unit_for(pack.objects[i])
        if unit and new_unit and (unit != new_unit): return ""
        unit = new_unit
    return unit
    
  def priority_for(self, pack):
    for i in range(len(pack.objects)):
      if self.name in pack.attrs[i]:
        priority = description_for_object(pack.objects[i])._get_attribute(self.name).priority_for(pack.objects[i])
        if priority: return priority
    return 0
    
  def field_class_for(self, pack, dont_get_value = False):
    field_class = None
    for i in range(len(pack.objects)):
      if self.name in pack.attrs[i]:
        new_field_class = description_for_object(pack.objects[i])._get_attribute(self.name).field_class_for(pack.objects[i], dont_get_value)
        if field_class and new_field_class and (not field_class is new_field_class): return None
        field_class = new_field_class
    return field_class
    
  def addable_values_to(self, o): return []

  def can_add_to(self, o): return False
  
  def can_remove_from(self, o):
    for i in range(len(o.objects)):
      if description_for_object(o.objects[i])._get_attribute(self.name).can_remove_from(o.objects[i]): return True
      
  def can_reorder(self, o):
    for i in range(len(o.objects)):
      if description_for_object(o.objects[i])._get_attribute(self.name).can_reorder(o.objects[i]): return True
      
  def get_value_for(self, pack):
    value = NonConsistent
    for i in range(len(pack.objects)):
      if self.name in pack.attrs[i]:
        v = description_for_object(pack.objects[i])._get_attribute(self.name).get_value_for(pack.objects[i])
        if value is NonConsistent: value = v
        elif value != v: return NonConsistent
    return value

  def set_value_for(self, pack, value):
    for i in range(len(pack.objects)):
      if self.name in pack.attrs[i]:
        description_for_object(pack.objects[i])._get_attribute(self.name).set_value_for(pack.objects[i], value)

  def __getitem__(self, index): return self

  
class ObjectPackDescription(object):
  attributes       = {} # For def_attr
  inherited_descrs = []
  is_owl_entity    = False
  def icon_filename_for(self, pack):
    icon_filenames = set()
    for o in pack.objects:
      icon_filenames.add(description_for_object(o).icon_filename_for(o))
    icon_filenames = sorted(icon_filenames)
    if len(icon_filenames) == 1: icon_filenames = icon_filenames * 2
    return icon_filenames
    
  def label_for   (self, pack): return editobj3.TRANSLATOR("(pack of %s objects)") % len(pack.objects)
  def details_for (self, pack): return ""
  
  def attributes_of(self, pack):
    attrs = set()
    for o_attrs in pack.attrs: attrs.update(o_attrs)
    attributes = [self._get_attribute(attr) for attr in attrs]
    return sorted(attributes, key = lambda attribute: (attribute.priority_for(pack), locale.strxfrm(attribute.label))) 
    
  def property_attributes_of(self, pack):
    attrs = set()
    for o_attrs in pack.property_attrs: attrs.update(o_attrs)
    return [self._get_attribute(attr) for attr in attrs]
    
  def _get_attribute(self, attr): return _object_pack_attribute(attr)
  
  def actions_for(self, parents, attributes, objects):
    if parents:        parents    = parents.objects
    else:              parents    = [None] * len(objects.objects)
    if not attributes: attributes = [None] * len(objects.objects)
    actions = set()
    for i in range(len(objects.objects)):
      actions.update(description_for_object(objects.objects[i]).actions_for(parents[i], attributes[i], objects.objects[i]))
    return actions
    
  def do_action(self, action, undo_stack, editor, pack, *args):
    if isinstance(action, ActionOnAChild):
      parents    = pack.objects
      attributes = args[0]
      objects    = args[1].objects
      if action.accept_object_pack:
        new_parents    = []
        new_attributes = []
        new_objects    = []
        for i in range(len(parents)):
          if action in description_for_object(objects[i]).actions_for(parents[i], attributes[i], objects[i]):
            new_parents   .append(parents   [i])
            new_attributes.append(attributes[i])
            new_objects   .append(objects   [i])
        return action.do(undo_stack, editor, ObjectPack(new_parents), new_attributes, ObjectPack(new_objects), *args[2:])
        
      else:
        for i in range(len(parents)):
          if action in description_for_object(objects[i]).actions_for(parents[i], attributes[i], objects[i]):
            description_for_object(parents[i]).do_action(action, undo_stack, parents[i], attributes[i], objects[i], *args[2:])
            
    else:
      if action.accept_object_pack:
        objects = ObjectPack([o for o in pack.objects if action in description_for_object(o).actions_for(None, None, o)])
        return action.do(undo_stack, editor, objects, *args)
        
      else:
        for o in pack.objects:
          descr = description_for_object(o)
          if action in descr.actions_for(None, None, o): descr.do_action(action, undo_stack, editor, o, *args)
        
_CLASS_DESCRS[ObjectPack] = ObjectPackDescription()



# -*- coding: utf-8 -*-
# Editobj3
# Copyright (C) 2007-2014 Jean-Baptiste LAMY

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from xml.sax.saxutils import escape
import gi
from gi.repository import GObject as gobject, Gdk as gdk, Gtk as gtk, GdkPixbuf as gdkpixbuf

import editobj3.introsp as introsp
from editobj3.editor import *

MENU_LABEL_2_IMAGE = {
  "Open..."        : gtk.STOCK_OPEN,
  "Save"           : gtk.STOCK_SAVE,
  "Save as..."     : gtk.STOCK_SAVE_AS,
  "Close"          : gtk.STOCK_CLOSE,
  "Undo"           : gtk.STOCK_UNDO,
  "Redo"           : gtk.STOCK_REDO,
  "Paste"          : gtk.STOCK_PASTE,
  "Preferences..." : gtk.STOCK_PREFERENCES,
  }

class GtkBaseDialog(BaseDialog):
  def on_response(self, dialog, response):
    if self.on_validate:
      if response == gtk.ResponseType.OK: self.on_validate(self.get_selected_object())
      else:                               self.on_validate(None)
    self.window.destroy()
      
  def build_default_menubar(self):
    self.accel_group = gtk.AccelGroup()
    self.window.add_accel_group(self.accel_group)
    self.current_radio_group = None
    
    self.menubar = gtk.MenuBar()
    self.window.get_child().pack_start(self.menubar, 0, 1, 0)
    
    BaseDialog.build_default_menubar(self)
    
  def add_to_menu(self, menu, has_submenu, label, command = None, arg = None, accel = "", accel_enabled = 1, image = None, type = u"button", pos = -1):
    if isinstance(menu, gtk.MenuItem):
      m = menu.get_submenu()
      if not m:
        m = gtk.Menu()
        menu.set_submenu(m)
      menu = m
    stock = MENU_LABEL_2_IMAGE.get(label)
    if stock: menu_item = gtk.ImageMenuItem.new_from_stock(stock, self.accel_group)
    elif image:
      gtk.stock_add([(label, editobj3.TRANSLATOR(label).replace("_", "__"), gtk.gdk.SHIFT_MASK, 46, "")])
      icon_factory.add(label, gtk.IconSet(image))
      menu_item = gtk.ImageMenuItem(label)
      self.current_radio_group = None
    elif type == "check":
      menu_item = gtk.CheckMenuItem(editobj3.TRANSLATOR(label).replace("_", "__"))
      self.current_radio_group = None
    elif type == "radio":
      menu_item = gtk.RadioMenuItem(self.current_radio_group, editobj3.TRANSLATOR(label).replace("_", "__"))
      self.current_radio_group = menu_item
    else:
      menu_item = gtk.MenuItem(editobj3.TRANSLATOR(label).replace("_", "__"))
      self.current_radio_group = None
    if command:
      if arg: menu_item.connect("activate", command, arg)
      else:   menu_item.connect("activate", command)
    menu.insert(menu_item, pos)
    
    if accel:
      mod = gdk.ModifierType(0)
      if isinstance(accel, int): key = accel
      else:
        if u"C-" in accel: mod |= gdk.ModifierType.CONTROL_MASK
        if u"S-" in accel: mod |= gdk.ModifierType.SHIFT_MASK
        key = ord(accel[-1])
      menu_item.add_accelerator("activate", self.accel_group, key, mod, gtk.AccelFlags.VISIBLE)
      if not accel_enabled:
        menu_item.connect("can-activate-accel", self.can_activate_accel)
    return menu_item
  
  def can_activate_accel(self, widget, signal):
    widget.stop_emission("can-activate-accel")
    return False
    
  def add_separator_to_menu(self, menu, pos = -1):
    menu.get_submenu().insert(gtk.SeparatorMenuItem(), pos)
    self.current_radio_group = None
    
  def set_menu_checked(self, menu, checked): menu.set_active(checked)
  def set_menu_enable (self, menu, enable):  menu.set_sensitive(enable)
  def set_menu_label  (self, menu, label):   menu.get_child().set_text(label)
  
  def on_dialog_closed(self, *args):
    self.window.destroy()
    if self.on_close: self.on_close()
    
  def main(self):
    self.show()
    gtk.main()
  
  def set_default_size(self, w, h): self.window.set_default_size(w, h)
    
  def show(self): self.window.show_all()


class GtkEditorDialog(EditorDialog, GtkBaseDialog):
  def __init__(self, gui = None, master = None, direction = "h", on_validate = None, edit_child_in_self = 1, undo_stack = None, on_close = None, menubar = True):
    if on_validate:
      flags = gtk.DialogFlags.MODAL | gtk.DialogFlags.DESTROY_WITH_PARENT | getattr(gtk.DialogFlags, "USE_HEADER_BAR", 0)
      buttons = (gtk.STOCK_CANCEL, gtk.ResponseType.CANCEL, gtk.STOCK_OK, gtk.ResponseType.OK)
      self.window = gtk.Dialog(None, None, flags, buttons)
      self.window.connect("response", self.on_response)
    else:
      self.window = gtk.Window(gtk.WindowType.TOPLEVEL)
      self.window.connect("delete_event", self.on_dialog_closed)
      self.box    = gtk.VBox()
      self.window.add(self.box)
    self.window.connect("delete_event", self.on_dialog_closed)
    
    super().__init__(gui, master, direction, on_validate, edit_child_in_self, undo_stack, on_close, menubar)
    self.window.get_child().pack_start(self.editor_pane, 1, 1, 0)
    self.window.set_default_size(_screen.width() - 128, _screen.height() - 64)
    
  def edit(self, o):
    EditorDialog.edit(self, o)
    label = self.editor_pane.hierarchy_pane.descr.label_for(o)
    self.window.set_title(label.replace("\n", " ").replace("\t", " "))
    return self
  

class GtkEditorTabbedDialog(EditorTabbedDialog, GtkBaseDialog):
  def __init__(self, gui = None, master = None, direction = "h", on_validate = None, edit_child_in_self = 1, undo_stack = None, on_close = None, menubar = True):
    if on_validate:
      flags = gtk.DialogFlags.MODAL | gtk.DialogFlags.DESTROY_WITH_PARENT | getattr(gtk.DialogFlags, "USE_HEADER_BAR", 0)
      buttons = (gtk.STOCK_CANCEL, gtk.ResponseType.CANCEL, gtk.STOCK_OK, gtk.ResponseType.OK)
      self.window = gtk.Dialog(None, None, flags, buttons)
      self.window.connect("response", self.on_response)
      self.window.show_all()
    else:
      self.window = gtk.Window(gtk.WindowType.TOPLEVEL)
      self.box    = gtk.VBox()
      self.window.add(self.box)
    self.window.connect("delete_event", self.on_dialog_closed)
      
    super().__init__(gui, master, direction, on_validate, edit_child_in_self, undo_stack, on_close, menubar)
    self.notebook = gtk.Notebook()
    self.window.get_child().pack_start(self.notebook, 1, 1, 0)
      
  def add_tab(self, name, label, o = None):
    editor_pane = super().add_tab(name, label, o)
    label = gtk.Label(label)
    label.set_alignment(1.0, 0.5)
    self.notebook.append_page(editor_pane, label)
    return editor_pane
    
  def remove_tab(self, name):
    for i in range(len(self.editor_panes)):
      if self.editor_panes[name] is self.notebook.get_nth_page(i):
        self.notebook.remove_page(i)
        break
    super.remove_tab(name)
    
  def get_current_editor_pane(self): return self.notebook.get_nth_page(self.notebook.get_current_page())
  
  
  
class GtkEditorPane(EditorPane, gtk.Paned):
  def __init__(self, gui, master, edit_child_in_self = 1, undo_stack = None, direction = "h"):
    gtk.Paned.__init__(self)
    if direction == "h": self.set_orientation(gtk.Orientation.HORIZONTAL)
    else:                self.set_orientation(gtk.Orientation.VERTICAL)
    super().__init__(gui, master, edit_child_in_self, undo_stack)
    
    #self.box = gtk.VBox()
    self.box = gtk.Box()
    self.box.set_orientation(gtk.Orientation.VERTICAL)
    self.box.pack_start(self.icon_pane     , 0, 1, 0)
    self.box.pack_start(self.attribute_pane, 0, 1, 0)
    self.scroll1 = gtk.ScrolledWindow()
    self.scroll2 = gtk.ScrolledWindow()
    self.scroll1.set_policy(gtk.PolicyType.AUTOMATIC, gtk.PolicyType.AUTOMATIC)
    self.scroll2.set_policy(gtk.PolicyType.NEVER    , gtk.PolicyType.AUTOMATIC)
    self.scroll1.set_shadow_type(gtk.ShadowType.IN)
    self.scroll1.add(self.hierarchy_pane)
    self.scroll2.add_with_viewport(self.box)
    self.scroll2.get_child().set_shadow_type(gtk.ShadowType.NONE)
    self.hi_box = gtk.Overlay()
    self.hi_box.add(self.scroll1)
    self.hi_box.add_overlay(self.childhood_pane)
    self.childhood_pane.set_property("halign", gtk.Align.END)
    self.childhood_pane.set_property("valign", gtk.Align.START)
    self.childhood_pane.set_margin_top(3)
    self.childhood_pane.set_margin_right(3)
    if direction == "h": self.pack1(self.hi_box , 1, 0)
    else:                self.pack1(self.hi_box , 0, 0)
    self.pack2(self.scroll2, 1, 0)
    
  def edit(self, o):
    EditorPane.edit(self, o)
    self.show_all()
    
  def _set_hierarchy_visible(self, visible):
    if visible:
      if self.get_position() == 0: self.set_position(500)
      self.scroll1.set_size_request(300, 200)
      self.child_set_property(self.hi_box, "shrink", 0)
    else:
      self.scroll1.set_size_request(0, 0)
      self.child_set_property(self.hi_box, "shrink", 1)
      def f(): self.set_position(0)
      gobject.timeout_add(0, f)
      
 
_screen = gdk.Screen()

class GtkAttributePane(AttributePane, gtk.Grid):
  def __init__(self, gui, master, edit_child = None, undo_stack = None):
    gtk.Grid.__init__(self)
    super().__init__(gui, master, edit_child, undo_stack)
    self.set_column_spacing(2)
    self.set_row_spacing(3)
    self.connect("destroy", self._destroyed)
    self.can_expand = 0
    
  def edit(self, o, ignored_attrs = None):
    if self.o is o: return
    self.can_expand = 0
    AttributePane.edit(self, o, ignored_attrs)
    
    if isinstance(self.master, EditorPane):
      if self.can_expand: self.master.box.set_child_packing(self, 1, 1, 0, gtk.PackType.START)
      else:               self.master.box.set_child_packing(self, 0, 1, 0, gtk.PackType.START)
      self.master.box.show_all()
      min_size, natural_size = self.master.box.get_preferred_size()
      if self.master.direction == "h": self.master.scroll2.set_min_content_height(min(natural_size.height, 800, _screen.height() - 128))
      else:                            self.master.scroll2.set_min_content_height(min(natural_size.height, 500, _screen.height() - 128))
    else: self.show_all()
    
  def _delete_all_fields(self):
    for widget in list(self.get_children()): widget.destroy()
    
  def _new_field(self, attribute, field_class, unit, i):
    field = field_class(self.gui, self, self.o, attribute, self.undo_stack)
    label = gtk.Label(attribute.label)
    label.set_alignment(0.0, 0.0)
    label.set_margin_top(5)
    self.attach(label, 0, i, 1, 1)
    self.attach(field, 1, i, 1, 1)
    field.set_hexpand(1)
    if unit:
      unit_label = gtk.Label(unit)
      unit_label.set_alignment(0.0, 0.5)
      self.attach(unit_label, 2, i, 1, 1)
    if field.y_flags & gtk.AttachOptions.EXPAND:
      field.set_vexpand(1)
      self.can_expand = 1
    return field
    

class GtkIconPane(IconPane, gtk.HBox):
  def __init__(self, gui, master):
    gtk.HBox.__init__(self)
    super().__init__(gui, master)
    
    self.image = None
    self.label = gtk.Label()
    self.label.set_line_wrap(1)
    self.label.set_alignment(0.0, 0.6)
    self.label.set_padding(10, 0)
    self.pack_end(self.label, 1, 1, 0)
    self.connect("destroy", self._destroyed)
    
  def _set_icon_filename_label_details(self, icon_filename, label, details):
    if   not icon_filename:
      if self.image:
        self.image.destroy()
        self.image = None
        
    elif isinstance(icon_filename, str):
      if isinstance(self.image, gtk.Fixed):
        self.image.destroy()
        self.image = None
      if not self.image:
        self.image = gtk.Image()
        self.image.set_alignment(0.5, 0.0)
        self.image.set_padding(10, 10)
        self.pack_start(self.image, 0, 1, 0)
        self.image.show_all()
      self.image.set_from_pixbuf(load_big_icon(icon_filename))
      
    else:
      if self.image: self.image.destroy()
      self.image = gtk.Fixed()
      self.image.set_border_width(5)
      x = y = 0
      icon_filename.sort()
      icon_filename.reverse()
      for filename in icon_filename:
        i = gtk.Image()
        i.set_from_pixbuf(load_big_icon(filename))
        self.image.put(i, x, y)
        x += 30
        y += 15
      self.pack_start(self.image, 0, 1, 0)
      self.image.show_all()
      
    if details: self.label.set_markup("<b>%s</b>\n\n%s\n" % (escape(label), escape(details)))
    else:       self.label.set_markup("<b>%s</b>"         %  escape(label))
    self.show_all()
  

class GtkChildhoodPane(ChildhoodPane, gtk.VBox):
  def __init__(self, gui, master, undo_stack = None, restrict_to_attribute = None):
    gtk.VBox.__init__(self)
    super().__init__(gui, master, undo_stack, restrict_to_attribute)
    
    self.button_move_up   = gtk.Button()
    self.button_add       = gtk.Button()
    self.button_remove    = gtk.Button()
    self.button_move_down = gtk.Button()
    self.button_move_up  .set_image(gtk.Image.new_from_stock(gtk.STOCK_GO_UP  , gtk.IconSize.BUTTON))
    self.button_add      .set_image(gtk.Image.new_from_stock(gtk.STOCK_ADD    , gtk.IconSize.BUTTON))
    self.button_remove   .set_image(gtk.Image.new_from_stock(gtk.STOCK_REMOVE , gtk.IconSize.BUTTON))
    self.button_move_down.set_image(gtk.Image.new_from_stock(gtk.STOCK_GO_DOWN, gtk.IconSize.BUTTON))
    self.button_move_up  .connect("clicked", self.on_move_up)
    self.button_add      .connect("clicked", self.on_add)
    self.button_remove   .connect("clicked", self.on_remove)
    self.button_move_down.connect("clicked", self.on_move_down)
    self.pack_start(self.button_move_up  , 0, 1, 0)
    self.pack_start(self.button_add      , 0, 1, 0)
    self.pack_start(self.button_remove   , 0, 1, 0)
    self.pack_start(self.button_move_down, 0, 1, 0)
    self.button_move_up  .set_relief(gtk.ReliefStyle.NONE)
    self.button_add      .set_relief(gtk.ReliefStyle.NONE)
    self.button_remove   .set_relief(gtk.ReliefStyle.NONE)
    self.button_move_down.set_relief(gtk.ReliefStyle.NONE)
    self.set_property("visible", 0)
    self.set_property("no-show-all", 1)
    
  def set_button_visibilities(self, visible1, visible2, visible3, visible4):
    if visible1 or visible2 or visible3 or visible4:
      self.button_move_up  .set_property("visible", visible1)
      self.button_add      .set_property("visible", visible2)
      self.button_remove   .set_property("visible", visible3)
      self.button_move_down.set_property("visible", visible4)
      self.button_move_up  .set_property("no-show-all", not visible1)
      self.button_add      .set_property("no-show-all", not visible2)
      self.button_remove   .set_property("no-show-all", not visible3)
      self.button_move_down.set_property("no-show-all", not visible4)
      self.set_property("visible", 1)
      self.set_property("no-show-all", 0)
      self.show_all()
    else:
      self.set_property("visible", 0)
      self.set_property("no-show-all", 1)
      

class DynamicNode(object):
  def __init__(self, parent):
    self.children         = []
    self.children_created = False
    self.is_expandable    = False
    self.index            = 0
    if isinstance(parent, DynamicNode):
      self.parent         = parent
      self.tree           = parent.tree
    else:
      self.parent         = None
      self.tree           = parent
      if not self.flat_list: self.tree.append(None)
      self.update()
      self.update_children()
      
  def path(self):
    if self.parent: return self.parent.path() + (self.index,)
    if self.flat_list: return ()
    else:              return (0,)
    
  def get_by_path(self, path, i = 0):
    if self.flat_list and not self.parent:
      pass
    else:
      if i == len(path) - 1: return self
      i += 1
    return self.children[path[i]].get_by_path(path, i)
    
  def has_children(self): raise NotImplementedError
  def create_children(self, old_children = None): return []
  
  def expand(self):
    path = self.path()
    if path: self.tree.view.expand_to_path(gtk.TreePath(path))
    
  def expanded(self):
    if not self.children_created:
      self.children_created = True
      self.update_children()
      if self.is_expandable: del self.tree[self.path() + (len(self.children),)]
    return self.children
    
  def collapsed(self):
    if self.children_created:
      self.children_created = False
      for child in self.children[::-1]: child.destroy()
      self.is_expandable = False
      self.update_children()
      
  def update(self): pass
  
  def update_children(self):
    if self.children_created:
      old_children = self.children[:]
      new_children = self.create_children(old_children)
      old_children_set = set(old_children)
      new_children_set = set(new_children)
      
      #print()
      #print(old_children, "=>", new_children)
      
      for child in old_children[::-1]:
        if not child in new_children_set: child.destroy()
        
      kept_children = [child for child in new_children if child in old_children_set]
      
      for i in range(len(self.children)): self.children[i].index     = i
      for i in range(len(kept_children)): kept_children[i].new_index = i
      
      if kept_children:
        for child in kept_children:
          if child.new_index == child.index : continue # Swap with itself
          swap_with = self.children[child.new_index]
          #print("  swap", child.path(), "with", swap_with.path())
          self.tree.swap(self.tree.get_iter(child.path()), self.tree.get_iter(swap_with.path()))
          swap_with.index = child.index
          child.index     = child.new_index
          self.children[swap_with.index] = swap_with
          self.children[child    .index] = child
          
        for child in kept_children: del child.new_index
        
      if self.flat_list: iter = None
      else:              iter = self.tree.get_iter(self.path())
      for i in range(len(new_children)):
        new_children[i].index = i
        if not new_children[i] in old_children_set:
          self.children.insert(i, new_children[i])
          self.tree.insert(iter, i)
          
      for child in new_children:
        if not child in old_children_set:
          child.update()
          child.update_children()
          
      self.children = new_children
      
    else:
      if self.flat_list: pass
      else:
        is_expandable = bool(self.has_children())
        if   is_expandable and (not self.is_expandable): self.tree.append(self.tree.get_iter(self.path()))
        elif self.is_expandable and (not is_expandable): del self.tree[self.path() + (0,)]
        self.is_expandable = is_expandable
        
  def destroy(self):
    for child in self.children[::-1]: child.destroy()
    path = self.path()
    if self.parent: self.parent.children.remove(self)
    if path and (self.tree.view.tree is self.tree): # Else, the TreeView is displaying a new tree.
      del self.tree[path]
    

class Gtk_HierarchyNode(HierarchyNode, DynamicNode):
  def update(self):
    icon_filename = self.descr.icon_filename_for(self.o)
    if isinstance(icon_filename, str): pixbuf = load_small_icon(icon_filename)
    else:                              pixbuf = None
    path = self.path()
    if path: self.tree[path] = str(self), pixbuf
    
  def __str__(self):
    if self.attribute and not self.restrict_to_attribute: attr_label = self.attribute.label
    else:                                                 attr_label = ""
    if attr_label: return '''<span size="smaller" weight="ultralight">%s</span> %s''' % (escape(attr_label), escape(self.descr.label_for(self.o)))
    return escape(self.descr.label_for(self.o))
    
    
class GtkHierarchyPane(HierarchyPane, gtk.TreeView):
  Node = Gtk_HierarchyNode
  def __init__(self, gui, master, edit_child, undo_stack = None, restrict_to_attribute = None, flat_list = 0):
    self.tree = gtk.TreeStore(str, gdkpixbuf.Pixbuf)
    self.tree.view = self
    gtk.TreeView.__init__(self, self.tree)
    column = gtk.TreeViewColumn(None)
    
    self.set_headers_visible(0)
    pixbuf_render = gtk.CellRendererPixbuf()
    column.pack_start(pixbuf_render, 0)
    column.add_attribute(pixbuf_render, "pixbuf", 1)
    text_render = gtk.CellRendererText()
    column.pack_start(text_render, 0)
    column.add_attribute(text_render, "markup", 0)
    self.append_column(column)
    
    super().__init__(gui, master, edit_child, undo_stack, restrict_to_attribute, flat_list)
    self.connect("destroy", self._destroyed)
    self.get_selection().set_mode(gtk.SelectionMode.MULTIPLE)
    self.get_selection().connect("changed", self.on_selection_changed)
    self.set_size_request(200, 200)
    self.selected_parent     = None
    self.selected_attribute  = None
    self.selected_node       = None
    self.selected            = None
    self.last_button         = 0
    
    #self.connect("row-expanded"      , self._expanded)
    self.connect("test-expand-row"   , self.on_expanded)
    self.connect("row-collapsed"     , self.on_collapsed)
    self.connect("button_press_event", self.on_button_pressed)
    
  def edit(self, o):
    if o is self.o: return
    
    old_tree = self.tree
    self.tree = gtk.TreeStore(str, gdkpixbuf.Pixbuf)
    self.tree.view = self
    self.set_model(None)
    old_tree.clear()
    
    HierarchyPane.edit(self, o)
    self.root_node.expanded()
    self.set_model(self.tree)
    if self.flat_list:
      self.childhood_pane.edit(None, None, self.o)
    else:
      self.expand_to_path((gtk.TreePath((0,))))
      self.select_node(self.root_node)
      #self.select_object(o)
    self.show_all()
    
  def expand_tree_at_level(self, level): self._expand_tree_at_level(level, (0,))
    
  def _expand_tree_at_level(self, level, path):
    self.expand_to_path(path)
    if len(path) < level:
      for i in range(self.tree.iter_n_children(self.tree[path].iter)):
        self._expand_tree_at_level(level, path + (i,))
        
  def select_node(self, node):
    self.get_selection().unselect_all()
    if node:
      path = node.path()
      if path: self.get_selection().select_path(path)
    elif self.flat_list and self.childhood_pane:
      self.childhood_pane.edit(None, None, self.o)
      
  def on_expanded(self, treeview, iter, path):
    node = self.root_node.get_by_path(path)
    node.expanded()
    self.handler_block_by_func(self.on_expanded)
    self.expand_to_path(path)
    self.handler_unblock_by_func(self.on_expanded)
    return True
    
  def on_collapsed(self, treeview, iter, path):
    node = self.root_node.get_by_path(path)
    node.collapsed()
    
  def on_button_pressed(self, widget, event):
    if event.type == gdk.EventType._2BUTTON_PRESS:
      actions = self.get_actions(self.selected_parent, self.selected_attribute, self.selected)
      for action in actions:
        if action.default:
          #self._action_activated(None, self.selected, action, self.selected_parent)
          self._action_activated(action, self.selected_parent, self.selected_attribute, self.selected)
          break
    else:
      self.last_button      = event.button
      self.last_button_time = event.time
      
  def on_selection_changed(self, *args):
    tree, rows = self.get_selection().get_selected_rows()
    if len(rows) == 0:
      #self.edit_child(None)
      delayed(lambda: self.edit_child(None)) # Delay because focus out event must be received by Field widget BEFORE edditing a new object!
      if self.childhood_pane: self.childhood_pane.edit(None, None, None)
      self.selected_node = None
      return
      
    if len(rows) == 1:
      node = self.selected_node = self.root_node.get_by_path(rows[0])
      if node.parent: self.selected_parent = node.parent.o
      else:           self.selected_parent = None
      self.selected           = node.o
      self.selected_attribute = node.attribute
      
      delayed(lambda: self.edit_child(self.selected, self.get_ignored_attrs([node]))) # Delay because focus out event must be received by Field widget BEFORE edditing a new object!
      
      if self.childhood_pane: self.childhood_pane.edit(self.selected_parent, self.selected_attribute, self.selected)
      
    else:
      nodes = self.selected_node = [self.root_node.get_by_path(row) for row in rows]
      self.selected_parent    = introsp.ObjectPack([(node.parent and node.parent.o) for node in nodes])
      self.selected           = introsp.ObjectPack([node.o for node in nodes])
      self.selected_attribute = [node.attribute for node in nodes]
      
      delayed(lambda: self.edit_child(self.selected, self.get_ignored_attrs(nodes))) # Delay because focus out event must be received by Field widget BEFORE edditing a new object!
      
      if self.childhood_pane: self.childhood_pane.edit(self.selected_parent, self.selected_attribute, self.selected)
      
    if self.last_button == 3:
      self.last_button = 0
      actions = self.get_actions(self.selected_parent, self.selected_attribute, self.selected)
      self.show_action_menu(actions, self.last_button_time)
      
  def on_action_activated(self, event, action): self._action_activated(action, self.selected_parent, self.selected_attribute, self.selected)
  
  def show_action_menu(self, actions, button_time = None):
    self._popup_menu = gtk.Menu() # The menu must be kept in memory, else it does not appear.
    #gtk.MenuItem("Test")
    for action in actions:
      menu_item = gtk.MenuItem(action.label)
      menu_item.connect("activate", self.on_action_activated, action)
      self._popup_menu.append(menu_item)
    self._popup_menu.show_all()
    self._popup_menu.popup(None, None, None, None, 0, button_time or gtk.get_current_event_time())


  
def delayed(f):
  gobject.timeout_add(0, f)
  return False
  
SMALL_ICON_SIZE = 32

_small_icons = {}
_big_icons   = {}

def load_big_icon(filename):
  pixbuf = _big_icons.get(filename)
  if not pixbuf: pixbuf = _big_icons[filename] = gdkpixbuf.Pixbuf.new_from_file(filename)
  return pixbuf

def load_small_icon(filename):
  pixbuf = _small_icons.get(filename)
  if not pixbuf:
    pixbuf = load_big_icon(filename)
    w = pixbuf.get_width()
    h = pixbuf.get_height()
    if (w > SMALL_ICON_SIZE) or (h > SMALL_ICON_SIZE):
      if w > h: pixbuf = pixbuf.scale_simple(SMALL_ICON_SIZE, int(float(SMALL_ICON_SIZE) * h / w), gdkpixbuf.InterpType.BILINEAR)
      else:     pixbuf = pixbuf.scale_simple(int(float(SMALL_ICON_SIZE) * w / h), SMALL_ICON_SIZE, gdkpixbuf.InterpType.BILINEAR)
    _small_icons[filename] = pixbuf
  return pixbuf



# -*- coding: utf-8 -*-
# Editobj3
# Copyright (C) 2007-2014 Jean-Baptiste LAMY

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, os

import PyQt5.QtCore    as qtcore
import PyQt5.QtWidgets as qtwidgets
import PyQt5.QtGui     as qtgui

import editobj3
import editobj3.introsp as introsp
import editobj3.observe as observe
import editobj3.field   as field
import editobj3.editor  as editor

editobj3.GUI = "Qt"

#app = qtwidgets.QApplication(sys.argv)

class Obj(object):
  def __init__(self, name, children = None):
    self.children = children or []
    self.children2 = []
    self.name = name
    self.prop = 200

  def __str__(self): return self.name
  
o = Obj("racine", [
  Obj("corneille", [
    Obj("molière"),
  ]),
  Obj("tolkien"),
])
o.text = "A\nB\nc"
o.aaa = 3
#o.children2.append(o.children[0])

descr = introsp.description(Obj)
descr.def_attr("children", field.ObjectListField)
descr.def_attr("children2", field.ObjectListField)

#window = qtwidgets.QWidget()
#layout = qtwidgets.QVBoxLayout()
#pane = editor.EditorPane("Qt", None)
#pane.edit(o)
#layout.addWidget(pane)
#window.setLayout(layout)
#window.show()

editobj3.edit(o).main()

for attr in o.__dict__:
  print(attr, repr(getattr(o, attr)))
#sys.exit(app.exec_())



# -*- coding: utf-8 -*-
# Editobj3
# Copyright (C) 2007-2014 Jean-Baptiste LAMY

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, os

import PyQt5.QtCore    as qtcore
import PyQt5.QtWidgets as qtwidgets

import editobj3
import editobj3.introsp as introsp
import editobj3.observe as observe
import editobj3.field   as field
import editobj3.editor  as editor

editobj3.GUI = "Qt"

app = qtwidgets.QApplication(sys.argv)

class Point(object):
  def __init__(self):
    self.x = 1
    self.y = 2
  def __repr__(self): return "Point(%s, %s)" % (self.x, self.y)
  
class Obj(object):
  def __init__(self):
    self.prop = 1
    self.prop_deux = 200
    self.texte = "Bla\nBla\nBla..."
    self.password = "abcd"
    self.filename = "/tmp/t.txt"
    self.url = "http://python.org"
    self.boo = True
    self.level = 1
    self.level2 = 2
    self.flags = ["A"]
    self.point = Point()
o = Obj()

descr = introsp.description(Obj)
descr.def_attr("prop", field.RangeField(0, 100))
descr.def_attr("level"           , field.EnumField({"low":0, "medium":1, "high":2, "paranoid":3}))
descr.def_attr("level2"          , field.EnumField(list(range(50))))
descr.def_attr("flags"           , field.EnumListField(["A", "B", "C", "D"]))
#descr.def_attr("point"           , field.ObjectAttributeField)

window = qtwidgets.QWidget()
layout = qtwidgets.QVBoxLayout()
attribute_pane = editor.AttributePane("Qt", None)
attribute_pane.edit(o)
layout.addWidget(attribute_pane)
window.setLayout(layout)
window.show()

app.exec_()
for attr in o.__dict__:
  print(attr, repr(getattr(o, attr)))
#sys.exit(app.exec_())



# -*- coding: utf-8 -*-
# Editobj3
# Copyright (C) 2007-2014 Jean-Baptiste LAMY

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, os, os.path, editobj3, editobj3.introsp as introsp, editobj3.observe as observe, editobj3.field as field, editobj3.undoredo as undoredo, editobj3.editor as editor

class User(object):
  def __init__(self, name):
    self.name          = name
    self.password      = "123"
    
  def __str__(self): return "%s %s" % (self.__class__.__name__, self.name)
    
class Admin(User):
  pass
    
class Connection(object):
  def __init__(self):
    self.type  = "modem"
    self.speed = 60
    
class Security(object):
  def __init__(self):
    self.level            = 2
    self.allow_javascript = 0
    self.block_popup      = 0
    self.enable_antispam  = 1
    self.enable_antispam  = 1
    

class Plugin(object):
  def __init__(self, name, **kargs):
    self.__dict__ = kargs
    self.name     = name
    self.active   = 1
    
  def __repr__(self): return "Plugin '%s'" % self.name
  
class Config(object):
  def __init__(self):
    self.connection     = Connection()
    self.security       = Security()
    self.users   = [
      User("Jiba"),
      ]
    self.plugins = [
      Plugin("GPG", command = "gpg"),
      Plugin("SVG"),
      Plugin("OggVorbis", driver_filename = "/dev/dsp")
      ]
    
    self.default_plugin = self.plugins[0]

config = Config()
# os.path.join(os.path.dirname(sys.argv[0]), "./jiba.png")


descr = introsp.description(User)
descr.set_icon_filename(os.path.join(os.path.dirname(sys.argv[0]), "./jiba.png"))
descr.def_attr("list_of_int", field.IntListField)
descr.set_constructor(introsp.Constructor(lambda klass, parent: klass("")))

introsp.def_attr("default_plugin", field.ObjectSelectorField, addable_values = lambda config: [introsp.NewInstanceOf(Plugin)] + config.plugins + [None])

descr = introsp.description(Connection)
descr.set_label("Connection")
descr.def_attr("type" , field.EnumField(["modem", "DSL", "ADSL"]))
descr.def_attr("speed", field.RangeField(0, 512), unit = "Ko/s")

descr = introsp.description(Security)
descr.set_label("Security")
descr.def_attr("level"           , field.EnumField({"low":0, "medium":1, "high":2, "paranoid":3}))
descr.def_attr("allow_javascript", field.BoolField)
descr.def_attr("block_popup"     , field.BoolField)

descr = introsp.description(Plugin)
descr.def_attr("active", field.BoolField)
descr.set_constructor(introsp.Constructor(lambda klass, parent: klass("")))

descr = introsp.description(Config)
descr.set_label("Configuration")
descr.def_attr("plugins", addable_values = [introsp.NewInstanceOf(Plugin)])
descr.def_attr("users"  , addable_values = [introsp.NewInstanceOf(User), introsp.NewInstanceOf(Admin)])

# The following are not needed because Editobj is smart enough to guess them;
# they are kept only for documentation purpose.

#descr.def_attr("identification", field.ObjectAttributeField)
#descr.def_attr("connection"    , field.ObjectAttributeField)
#descr.def_attr("security"      , field.ObjectAttributeField)
#descr.set_label(lambda o: unicode(o))


if len(sys.argv) > 1: editobj3.GUI = sys.argv[-1]

editobj3.edit(config, on_close = sys.exit).main()




# -*- coding: utf-8 -*-
# Editobj3
# Copyright (C) 2007-2014 Jean-Baptiste LAMY

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import gi
from gi.repository import GObject as gobject, Gdk as gdk, Gtk as gtk, GdkPixbuf as gdkpixbuf

import editobj3
from editobj3.field import *
from editobj3.field import _WithButtonField, _RangeField, _ShortEnumField, _LongEnumField, _EnumListField

def _edit_from_field(o, widget, undo_stack): # GTK does not like that we show a non-modal dialog box when a modal dialog box is displayed
  p = widget
  while p:
    p = p.get_parent()
    if isinstance(p, gtk.Window) and p.get_modal():
      editobj3.edit(o, undo_stack = undo_stack, on_validate = lambda obj: None)
      return
  editobj3.edit(o, undo_stack = undo_stack)

class GtkField(MultiGUIField):
  y_flags = gtk.AttachOptions.FILL

class GtkHiddenField(GtkField, HiddenField): pass
  
class GtkLabelField(GtkField, LabelField, gtk.Label):
  def __init__(self, gui, master, o, attribute, undo_stack):
    gtk.Label.__init__(self)
    super().__init__(gui, master, o, attribute, undo_stack)
    self.update()
    
  def update(self):
    self.set_text(self.get_value())
    
class GtkEntryField(GtkField, EntryField, gtk.Entry):
  def __init__(self, gui, master, o, attribute, undo_stack):
    gtk.Entry.__init__(self)
    super().__init__(gui, master, o, attribute, undo_stack)
    
    self.update()
    self.connect("focus_out_event", self.validate)
    self.connect("key_press_event", self.validate)
    
  def validate(self, widget, event):
    if (event.type == gdk.EventType.KEY_PRESS) and ((not event.string) or (not event.string in "\r\n")): return
    self.set_value(self.get_text())
    
  def update(self):
    self.updating += 1
    try: self.set_text(self.get_value() or "") # GTK does not accept None as a text
    finally: self.updating -= 1

class GtkTextField(GtkField, TextField, gtk.ScrolledWindow):
  y_flags = gtk.AttachOptions.FILL | gtk.AttachOptions.EXPAND
  def __init__(self, gui, master, o, attribute, undo_stack):
    gtk.ScrolledWindow.__init__(self)
    self.set_policy(gtk.PolicyType.AUTOMATIC, gtk.PolicyType.AUTOMATIC)
    self.set_shadow_type(gtk.ShadowType.IN)
    self.set_size_request(-1, 125)
    self.text = gtk.TextView()
    self.text.set_wrap_mode(gtk.WrapMode.WORD)
    self.text.set_size_request(150, -1)
    self.text.connect("focus_out_event", self.validate)
    self.add(self.text)
    super().__init__(gui, master, o, attribute, undo_stack)
    self.update()
    
  def validate(self, *args):
    buffer = self.text.get_buffer()
    s = buffer.get_text(buffer.get_start_iter(), buffer.get_end_iter(), 1)
    self.set_value(s)
    
  def update(self):
    self.updating += 1
    try:
      value = self.get_value()
      buffer = self.text.get_buffer()
      if buffer.get_text(buffer.get_start_iter(), buffer.get_end_iter(), 1) != value: buffer.set_text(value)
    finally: self.updating -= 1
    
    
class GtkIntField   (GtkEntryField, IntField   ): pass # XXX no "spin-button" since they don't allow entering e.g. "1 + 2" as an integer !
class GtkFloatField (GtkEntryField, FloatField ): pass
class GtkStringField(GtkEntryField, StringField): pass

class GtkPasswordField(GtkStringField, PasswordField):
  def __init__(self, gui, master, o, attribute, undo_stack):
    GtkStringField.__init__(self, gui, master, o, attribute, undo_stack)
    self.set_visibility(0)


class GtkEntryListField (GtkTextField, EntryListField ): pass
class GtkIntListField   (GtkTextField, IntListField   ): pass
class GtkFloatListField (GtkTextField, FloatListField ): pass
class GtkStringListField(GtkTextField, StringListField): pass
    
class Gtk_WithButtonField(GtkField, _WithButtonField, gtk.HBox):
  def __init__(self, gui, master, o, attribute, undo_stack, field_class = None, button_text = None, on_button = None):
    gtk.HBox.__init__(self)
    super().__init__(gui, master, o, attribute, undo_stack, field_class, button_text, on_button)
    self.pack_start(self.string_field, 1, 1, 0)
    button = gtk.Button(editobj3.TRANSLATOR(self.button_text))
    button.connect("clicked", self.on_button_clicked)
    self.pack_end(button, 0, 1, 0)
  
  def on_button_clicked(self, *args): self.on_button(self.o, self, self.undo_stack)
    
class GtkFilenameField(Gtk_WithButtonField, FilenameField):
  def on_button(self, o, field, undo_stack):
    dialog = gtk.FileChooserDialog(action = gtk.FileChooserAction.SAVE, buttons = (gtk.STOCK_CANCEL, gtk.ResponseType.CANCEL, gtk.STOCK_OK, gtk.ResponseType.OK))
    dialog.set_resizable(1)
    dialog.set_current_name(self.get_value())
    if dialog.run() == gtk.ResponseType.OK:
      filename = dialog.get_filename()
      if filename:
        self.string_field.set_value(filename)
        self.string_field.update()
    dialog.destroy()
    
class GtkDirnameField(Gtk_WithButtonField, DirnameField):
  def on_button(self, o, field, undo_stack):
    dialog = gtk.FileChooserDialog(action = gtk.FileChooserAction.CREATE_FOLDER, buttons = (gtk.STOCK_CANCEL, gtk.ResponseType.CANCEL, gtk.STOCK_OK, gtk.ResponseType.OK))
    dialog.set_resizable(1)
    dialog.set_current_folder(self.get_value())
    if dialog.run() == gtk.ResponseType.OK:
      folder = dialog.get_current_folder()
      if folder:
        self.string_field.set_value(folder)
        self.string_field.update()
    dialog.destroy()
    
class GtkURLField(Gtk_WithButtonField, URLField): pass


class GtkBoolField(GtkField, BoolField, gtk.CheckButton):
  def __init__(self, gui, master, o, attribute, undo_stack):
    gtk.CheckButton.__init__(self)
    self.connect("toggled", self.validate)
    self.connect("clicked", self.clicked)
    super().__init__(gui, master, o, attribute, undo_stack)
    self.update()
    
  def validate(self, *event): self.set_value(self.get_active())
    
  def clicked(self, *event):
    if self.get_inconsistent(): self.set_inconsistent(0)
    
  def update(self):
    self.updating += 1
    try:
      v = self.get_value()
      if v is introsp.NonConsistent: self.set_inconsistent(1)
      else: self.set_active(v)
    finally: self.updating -= 1


class GtkProgressBarField(GtkField, ProgressBarField, gtk.ProgressBar):
  def __init__(self, gui, master, o, attribute, undo_stack):
    gtk.ProgressBar.__init__(self)
    super().__init__(gui, master, o, attribute, undo_stack)
    self.update()
    
  def update(self):
    v = self.get_value()
    if v is introsp.NonConsistent: self.pulse()
    else: self.set_fraction(v)


class Gtk_RangeField(GtkField, _RangeField, gtk.HScale):
  def __init__(self, gui, master, o, attribute, undo_stack, min, max, incr = 1):
    gtk.HScale.__init__(self)
    self.adjustment = gtk.Adjustment(0, min, max, incr)
    self.set_adjustment(self.adjustment)
    self.set_digits(0)
    super().__init__(gui, master, o, attribute, undo_stack, min, max, incr)
    self.connect("value_changed", self.validate)
    
  def validate(self, *args): self.set_value(self.adjustment.get_value())
  
  def update(self):
    self.updating += 1
    try: self.adjustment.set_value(self.get_value())
    finally: self.updating -= 1
    
    
class Gtk_ShortEnumField(GtkField, _ShortEnumField, gtk.ComboBox):
  def __init__(self, gui, master, o, attribute, undo_stack, choices, value_2_enum = None, enum_2_value = None):
    gtk.ComboBox.__init__(self)
    self.liststore = gtk.ListStore(str)
    self.set_model(self.liststore)
    cell = gtk.CellRendererText()
    self.pack_start(cell, 1)
    self.add_attribute(cell, 'text', 0)
    super().__init__(gui, master, o, attribute, undo_stack, choices, value_2_enum, enum_2_value)
    for choice in self.choice_keys: self.liststore.append((choice,))
    self.connect("changed", self.validate)
    self.update()
    
  def validate(self, *args):
    i = self.get_active()
    if i != -1: self.set_value(self.choices[self.choice_keys[i]])
    
  def update(self):
    self.updating += 1
    try:
      i = self.choice_2_index.get(self.get_value())
      if i is None: self.set_active(-1)
      else:         self.set_active(i)
    finally: self.updating -= 1
    
class Gtk_LongEnumField(GtkField, _LongEnumField, gtk.ScrolledWindow):
  y_flags = gtk.AttachOptions.FILL | gtk.AttachOptions.EXPAND
  def __init__(self, gui, master, o, attribute, undo_stack, choices, value_2_enum = None, enum_2_value = None):
    gtk.ScrolledWindow.__init__(self)
    self.set_policy(gtk.PolicyType.AUTOMATIC, gtk.PolicyType.AUTOMATIC)
    self.set_shadow_type(gtk.ShadowType.IN)
    self.set_size_request(-1, 125)
    
    super().__init__(gui, master, o, attribute, undo_stack, choices, value_2_enum, enum_2_value)
    
    self.liststore = gtk.ListStore(str)
    for choice in self.choice_keys: self.liststore.append((choice,))
    renderer = gtk.CellRendererText()
    self.treeview = gtk.TreeView(self.liststore)
    self.treeview.set_headers_visible(0)
    self.treeview.append_column(gtk.TreeViewColumn(None, renderer, text = 0))
    self.add(self.treeview)
    
    self.update()
    self.treeview.get_selection().connect("changed", self.validate)
    
  def validate(self, *args):
    liststore, iter = self.treeview.get_selection().get_selected()
    if iter:
      enum = self.choices[self.choice_keys[int(liststore.get_path(iter)[0])]]
      self.set_value(enum)
      
  def update(self):
    self.updating += 1
    try:
      selection = self.treeview.get_selection()
      selection.unselect_all()
      self.i = self.choice_2_index.get(self.get_value())
      if not self.i is None:
        selection.select_iter(self.liststore.get_iter(self.i))
        self.treeview.scroll_to_cell(self.i)
    finally: self.updating -= 1
    
class Gtk_EnumListField(GtkField, _EnumListField, gtk.ScrolledWindow):
  y_flags = gtk.AttachOptions.FILL | gtk.AttachOptions.EXPAND
  def __init__(self, gui, master, o, attribute, undo_stack, choices, value_2_enum = None, enum_2_value = None):
    gtk.ScrolledWindow.__init__(self)
    self.set_policy(gtk.PolicyType.AUTOMATIC, gtk.PolicyType.AUTOMATIC)
    self.set_shadow_type(gtk.ShadowType.IN)
    self.set_size_request(-1, 125)
    
    super().__init__(gui, master, o, attribute, undo_stack, choices, value_2_enum, enum_2_value)
    self.liststore = gtk.ListStore(bool, str)
    for choice in self.choice_keys: self.liststore.append((False, choice))
    renderer1 = gtk.CellRendererToggle()
    renderer2 = gtk.CellRendererText  ()
    self.treeview = gtk.TreeView(self.liststore)
    self.treeview.set_headers_visible(0)
    self.treeview.append_column(gtk.TreeViewColumn(None, renderer1, active = 0))
    self.treeview.append_column(gtk.TreeViewColumn(None, renderer2, text   = 1))
    self.add(self.treeview)
    
    self.update()
    renderer1.connect("toggled", self.on_cell_toggled, self.liststore)
    
  def on_cell_toggled(self, widget, path, model):
    values = self.get_value()
    value  = self.choices[self.choice_keys[int(path)]]
    if values is introsp.NonConsistent: return # XXX does not support yet edition of multiple object
    if model[path][0]:
      values.remove(value)
      model[path][0] = False
    else:
      if isinstance(values, set): values.add   (value)
      else:                       values.append(value)
      model[path][0] = True
      
  def update(self):
    self.updating += 1
    try:
      values = self.get_value()
      if values is introsp.NonConsistent:
        for i in range(len(self.choices)): self.liststore[i][0] = False
      else:
        if values: self.indexes = { self.choice_2_index.get(value) for value in values }
        else:      self.indexes = set()
        for i in range(len(self.choices)): self.liststore[i][0] = i in self.indexes
    finally: self.updating -= 1

    
class GtkObjectAttributeField(GtkField, ObjectAttributeField, gtk.Frame):
  def __init__(self, gui, master, o, attribute, undo_stack):
    super().__init__(gui, master, o, attribute, undo_stack)
    gtk.Frame.__init__(self)
    self.set_shadow_type(gtk.ShadowType.IN)
    self.attribute_pane.set_property("border-width", 2)
    self.add(self.attribute_pane)
    
class GtkObjectSelectorField(GtkField, ObjectSelectorField, gtk.HBox):
  def __init__(self, gui, master, o, attribute, undo_stack):
    gtk.HBox.__init__(self)
    self.current_addable_values = None
    self.combo = gtk.ComboBox()
    space_render  = gtk.CellRendererText()
    pixbuf_render = gtk.CellRendererPixbuf()
    text_render   = gtk.CellRendererText()
    self.combo.pack_start(space_render , 0)
    self.combo.pack_start(pixbuf_render, 0)
    self.combo.pack_start(text_render  , 0)
    self.combo.add_attribute(space_render , 'text'  , 2)
    self.combo.add_attribute(pixbuf_render, 'pixbuf', 1)
    self.combo.add_attribute(text_render  , 'text'  , 0)
    self.button = gtk.Button(editobj3.TRANSLATOR("Edit..."))
    self.button.connect("clicked", self.on_click)
    self.pack_start(self.combo , 1, 1, 0)
    self.pack_end  (self.button, 0, 1, 0)
    
    super().__init__(gui, master, o, attribute, undo_stack)
    
    self.combo.connect("changed", self.on_change_object)
    
  def destroy(self):
    super().destroy()
    observe.unobserve(self.get_value(), self._value_listener)
    
  def on_click(self, *args): _edit_from_field(self.get_value(), self, self.undo_stack)
  
  def on_change_object(self, *args):
    self.set_value(self.current_addable_values[self.combo.get_active()])
  
  def store_item_for(self, obj):
    descr = introsp.description_for_object(obj)
    icon_filename = descr.icon_filename_for(obj)
    if isinstance(icon_filename, str): pixbuf = editobj3.editor_gtk.load_small_icon(icon_filename)
    else:                              pixbuf = None
    return (descr.label_for(obj), pixbuf, " ")
    
  def _value_listener(self, obj, type, new, old):
    self.liststore[self.combo.get_active()] = self.store_item_for(obj)
    
  def update(self):
    self.updating += 1
    try:
      current_addable_values = self.get_addable_values()
      if current_addable_values != self.current_addable_values:
        self.current_addable_values = current_addable_values
        self.liststore = gtk.ListStore(str, gdkpixbuf.Pixbuf, str)
        for i in self.current_addable_values: self.liststore.append(self.store_item_for(i))
        
        self.combo.set_model(self.liststore)
      value = self.get_value()
      self.combo.set_active(self.current_addable_values.index(value))
    finally: self.updating -= 1
    

class GtkObjectListField(GtkField, ObjectListField, gtk.Overlay):
  y_flags = gtk.AttachOptions.FILL | gtk.AttachOptions.EXPAND
  def __init__(self, gui, master, o, attribute, undo_stack):
    gtk.Overlay.__init__(self)
    super().__init__(gui, master, o, attribute, undo_stack)
    self.scroll = gtk.ScrolledWindow()
    self.scroll.set_min_content_height(136) # Required for childhood pane
    self.scroll.set_policy(gtk.PolicyType.AUTOMATIC, gtk.PolicyType.AUTOMATIC)
    self.scroll.set_shadow_type(gtk.ShadowType.IN)
    self.scroll.add(self.hierarchy_pane)
    self.add(self.scroll)
    self.add_overlay(self.childhood_pane)
    self.childhood_pane.set_property("halign", gtk.Align.END)
    self.childhood_pane.set_property("valign", gtk.Align.START)
    self.childhood_pane.set_margin_top(3)
    self.childhood_pane.set_margin_right(3)
    self.hierarchy_pane.connect("focus_out_event", self.on_focus_out)
    
  def on_focus_out(self, *args):
    self.timer = gobject.timeout_add(200, self.hide_selection)
  
  def hide_selection(self):
    try:
      if self.childhood_pane.button_move_up.has_focus() or self.childhood_pane.button_move_down.has_focus() or self.childhood_pane.button_add.has_focus() or self.childhood_pane.button_remove.has_focus():
        return True
      else:
        self.hierarchy_pane.select_node(None)
        return False
    except: return False
    

#class GtkHierarchyAndObjectAttributeField(GtkObjectAttributeField, HierarchyAndObjectAttributeField): pass
#class GtkHierarchyAndObjectSelectorField(GtkObjectSelectorField, HierarchyAndObjectSelectorField): pass
#class GtkHierarchyAndObjectListField(GtkObjectListField, HierarchyAndObjectListField): pass



# -*- coding: utf-8 -*-
# Editobj3
# Copyright (C) 2007-2014 Jean-Baptiste LAMY

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os.path
from xml.sax.saxutils import escape

import PyQt5.QtCore    as qtcore
import PyQt5.QtWidgets as qtwidgets
import PyQt5.QtGui     as qtgui

import editobj3.introsp as introsp
from editobj3.editor import *

MENU_LABEL_2_IMAGE = {
  "Open..."        : "document-open",
  "Save"           : "document-save",
  "Save as..."     : "document-save-as",
  "Close"          : "window-close",
  "Undo"           : "edit-undo",
  "Redo"           : "edit-redo",
  "Paste"          : "edit-paste",
  "Preferences..." : "preferences-desktop",
  }

_KEEP = set()
def delayed(func, *args):
  def on_timer():
    func(*args)
    _KEEP.discard(timer)
  timer = qtcore.QTimer()
  timer.timeout.connect(on_timer)
  timer.setInterval(0)
  timer.setSingleShot(True)
  timer.start()
  _KEEP.add(timer)
  return timer

class QtBaseDialog(BaseDialog):
  def init_window(self, on_validate, master = None):
    if on_validate:
      self.window = qtwidgets.QDialog()
      self.window.setModal(True)
      self.layout = qtwidgets.QVBoxLayout()
      self.layout.setContentsMargins(0, 0, 0, 0)
      self.layout.setSpacing(0)
      
      self.cancel_button = qtwidgets.QPushButton(editobj3.TRANSLATOR("Cancel"))
      self.ok_button     = qtwidgets.QPushButton(editobj3.TRANSLATOR("Ok"))
      self.cancel_button.clicked.connect(lambda : self.window.done(0))
      self.ok_button    .clicked.connect(lambda : self.window.done(1))
      self.cancel_button.setIcon(qtgui.QIcon.fromTheme("dialog-cancel"))
      self.ok_button    .setIcon(qtgui.QIcon.fromTheme("dialog-ok"))
      self.cancel_button.setAutoDefault(False)
      self.ok_button    .setDefault(True)
      
      button_layout = qtwidgets.QHBoxLayout()
      button_layout.setContentsMargins(5, 5, 5, 5)
      button_layout.setSpacing(20)
      button_layout.addStretch(1)
      button_layout.addWidget(self.cancel_button)
      button_layout.addWidget(self.ok_button)
      
      self.layout.addLayout(button_layout)
      self.window.setLayout(self.layout)
    else:
      if master:
        if isinstance(master, QtBaseDialog): master = master.window
        self.window = qtwidgets.QDialog(master)
      else:
        self.window = qtwidgets.QWidget()
      self.window.closeEvent = self.closeEvent
      self.layout = qtwidgets.QVBoxLayout()
      self.layout.setContentsMargins(0, 0, 0, 0)
      self.layout.setSpacing(0)
      self.window.setLayout(self.layout)
      
  def build_default_menubar(self):
    self._menus = []
    self.current_radio_group = None
    self.menubar = qtwidgets.QMenuBar()
    self.layout.insertWidget(0, self.menubar)
    
    BaseDialog.build_default_menubar(self)
    
  def add_to_menu(self, menu, has_submenu, label, command = None, arg = None, accel = "", accel_enabled = True, image = None, type = u"button", pos = -1):
    if command:
      if arg: command2 = lambda *args: command(*(list(args) + [arg]))
      else:   command2 = command
      
    stock = MENU_LABEL_2_IMAGE.get(label)
    if has_submenu:
      submenu = qtwidgets.QMenu(editobj3.TRANSLATOR(label))
      self._menus.append(submenu)
      if pos == -1: menu.addMenu(submenu)
      else:         menu.insertMenu(menu.actions()[pos], submenu)
      if command: submenu.aboutToShow.connect(command2)
      return submenu
    
    else:
      action = qtwidgets.QAction(editobj3.TRANSLATOR(label), self.window)
      self._menus.append(action)
      if pos == -1: menu.addAction(action)
      else:         menu.insertAction(menu.actions()[pos], action)
      
      if   type == "check":
        action.setCheckable(True)
        self.current_radio_group = None
        if command: action.toggled.connect(command2)
      
      elif type == "radio":
        action.setCheckable(True)
        if not self.current_radio_group: self.current_radio_group = qtwidgets.QActionGroup(self.window)
        self.current_radio_group.addAction(action)
        if command: action.toggled.connect(command2)
        
      else:
        self.current_radio_group = None
        if isinstance(stock, qtgui.QIcon):                              image = stock
        elif stock and (not image) and qtgui.QIcon.hasThemeIcon(stock): image = qtgui.QIcon.fromTheme(stock)
        if image: action.setIcon(image)
        if command: action.triggered.connect(command2)
        
      if accel:
        accel = accel.replace("C-", "Ctrl+").replace("S-", "Shift+")
        action.setShortcut(qtgui.QKeySequence(accel))
        if not accel_enabled: action.setShortcutContext(qtcore.Qt.WidgetShortcut)
      return action
    
  def add_separator_to_menu(self, menu, pos = -1):
    if pos == -1: menu.addSeparator()
    else:         menu.insertSeparator(menu.actions()[pos])
    self.current_radio_group = None
    
  def set_menu_checked(self, menu, checked): menu.setChecked(checked)
  def set_menu_enable (self, menu, enable):  menu.setEnabled(enable)
  def set_menu_label  (self, menu, label):
    if isinstance(menu, qtwidgets.QMenu): menu.setTitle(label)
    else:                                 menu.setText (label)
    
  def close_dialog(self): self.window.close()
  def closeEvent(self, event):
    if self.on_dialog_closed(): event.accept()
    else:                       event.ignore()
  def on_dialog_closed(self, *args):
    if self.on_close: self.on_close()
    return True
  
  def main(self):
    self.show()
    return qtcore.QCoreApplication.exec_()
  
  def set_default_size(self, w, h): self.window.resize(w, h)
  
  def show(self):
    if self.on_validate:
      r = self.window.exec_()
      if r: self.on_validate(self.get_selected_object())
      else: self.on_validate(None)
      self.window.destroy()
    else: self.window.show()
  

class QtEditorDialog(EditorDialog, QtBaseDialog):
  def __init__(self, gui = None, master = None, direction = "h", on_validate = None, edit_child_in_self = 1, undo_stack = None, on_close = None, menubar = True):
    self.init_window(on_validate, master)
    super().__init__(gui, master, direction, on_validate, edit_child_in_self, undo_stack, on_close, menubar)
    if menubar: self.layout.insertWidget(1, self.editor_pane, 1)
    else:       self.layout.insertWidget(0, self.editor_pane, 1)
      
  def edit(self, o):
    EditorDialog.edit(self, o)
    label = self.editor_pane.hierarchy_pane.descr.label_for(o)
    self.window.setWindowTitle(label.replace("\n", " ").replace("\t", " "))
    return self


class QtEditorTabbedDialog(EditorTabbedDialog, QtBaseDialog):
  def __init__(self, gui = None, master = None, direction = "h", on_validate = None, edit_child_in_self = 1, undo_stack = None, on_close = None, menubar = True):
    self.init_window(on_validate, master)
    super().__init__(gui, master, direction, on_validate, edit_child_in_self, undo_stack, on_close, menubar)
    self.notebook = qtwidgets.QTabWidget()
    self.notebook.currentChanged.connect(self.on_tab_changed)
    if menubar: self.layout.insertWidget(1, self.notebook, 1)
    else:       self.layout.insertWidget(0, self.notebook, 1)

  # For updating the childhood pane layout
  def on_tab_changed(self, index):
    editor_pane = self.notebook.widget(index)
    self.show_tab(editor_pane)
    editor_pane.hi_box.layout().invalidate()
    
  def show(self):
    QtBaseDialog.show(self)
    editor_pane = self.notebook.currentWidget()
    if editor_pane:
      self.show_tab(editor_pane)
      editor_pane.hi_box.layout().invalidate()
      
  def add_tab(self, name, label, o = None, func = None):
    editor_pane = super().add_tab(name, label, o)
    editor_pane.func = func
    self.notebook.addTab(editor_pane, label)
    return editor_pane
  
  def remove_tab(self, name):
    for i in range(len(self.editor_panes)):
      if self.editor_panes[name] is self.notebook.widget(i):
        self.notebook.removeTab(i)
        break
    super.remove_tab(name)
    
  def get_current_editor_pane(self): return self.notebook.currentWidget()


class OverlayLayout(qtwidgets.QLayout):
  def __init__(self):
    qtwidgets.QLayout.__init__(self)
    self._items = []
    self.overlay_offset_x = 3
    self.overlay_offset_y = 4
    
  def count(self): return len(self._items)
  def addItem(self, item): self._items.append(item)
  def itemAt(self, i):
    if 0 <= i < len(self._items): return self._items[i]
  def takeAt(self, i):
    if 0 <= i < len(self._items):
      item = self._items[i]
      del self._items[i]
      return item
    
  def setGeometry(self, rect):
    self._items[0].setGeometry(rect)
    
    def set_geom1():
      size = self._items[1].sizeHint()
      size2 = self._items[0].widget().viewport().size()
      self._items[1].setGeometry(qtcore.QRect(rect.left() + size2.width() - size.width() - self.overlay_offset_x, rect.top() + self.overlay_offset_y, size.width(), size.height()))
    delayed(set_geom1)
    
  def sizeHint(self): return self._items[0].sizeHint()
  def minimumSize(self): return self._items[0].minimumSize()

class QtEditorPane(EditorPane, qtwidgets.QSplitter):
  def __init__(self, gui, master, edit_child_in_self = 1, undo_stack = None, direction = "h"):
    qtwidgets.QSplitter.__init__(self)
    if direction == "h": self.setOrientation(qtcore.Qt.Horizontal)
    else:                self.setOrientation(qtcore.Qt.Vertical)
    super().__init__(gui, master, edit_child_in_self, undo_stack)
    
    self.box = qtwidgets.QWidget()
    layout = qtwidgets.QVBoxLayout()
    layout.setContentsMargins(3, 3, 3, 3)
    layout.addWidget(self.icon_pane)
    layout.addWidget(self.attribute_pane)
    self.box.setLayout(layout)
    self.scroll2 = qtwidgets.QScrollArea()
    self.scroll2.setWidgetResizable(True)
    self.scroll2.setSizeAdjustPolicy(qtwidgets.QAbstractScrollArea.AdjustToContents)
    self.scroll2.setHorizontalScrollBarPolicy(qtcore.Qt.ScrollBarAlwaysOff)
    
    self.hi_box = qtwidgets.QWidget()
    layout2 = OverlayLayout()
    layout2.addWidget(self.hierarchy_pane)
    layout2.addWidget(self.childhood_pane)
    self.hi_box.setLayout(layout2)
    self.addWidget(self.hi_box)
    self.addWidget(self.scroll2)
    self.setCollapsible(1, False)
    
  def edit(self, o):
    EditorPane.edit(self, o)
    
    self.scroll2.setWidget(None)
    self.scroll2.setWidget(self.box)
    
    width = self.box.sizeHint().width()
    max_width = qtwidgets.QApplication.desktop().screenGeometry(self).width() - 64
    if not self.isCollapsible(0): max_width = max_width - self.hi_box.sizeHint().width() - self.handleWidth()
    if width <= max_width:
      self.scroll2.setMinimumWidth(width)
      self.scroll2.setHorizontalScrollBarPolicy(qtcore.Qt.ScrollBarAlwaysOff)
    else:
      self.scroll2.setMinimumWidth(max_width)
      self.scroll2.setHorizontalScrollBarPolicy(qtcore.Qt.ScrollBarAsNeeded)

  def _set_hierarchy_visible(self, visible):
    if visible:
      if self.isCollapsible(0):
        self.setCollapsible(0, False)
        self.setSizes([1, 1])
    else:
      self.setCollapsible(0, True)
      self.setSizes([0, 1])
      
  def sizeHint(self):
    if self.isCollapsible(0):
      size = qtcore.QSize(self.scroll2.sizeHint())
      size.setWidth (size.width()  + self.handleWidth())
      size.setHeight(size.height() + 40) # For ObjectPack edition => bigger icon pane
      return size
    else: return qtwidgets.QSplitter.sizeHint(self)
      

class QtAttributePane(AttributePane, qtwidgets.QWidget):
  def __init__(self, gui, master, edit_child = None, undo_stack = None):
    qtwidgets.QWidget.__init__(self)
    super().__init__(gui, master, edit_child, undo_stack)
    self.destroyed.connect(self._destroyed)
    self.can_expand = 0
    self._re_edit_timer = None
    
    self.grid_layout = qtwidgets.QGridLayout()
    self.grid_layout.setHorizontalSpacing(2)
    self.grid_layout.setVerticalSpacing(3)
    self.grid_layout.setContentsMargins(0, 0, 0, 0)
    self.setLayout(self.grid_layout)
    
  def _re_edit(self):
    if not self._re_edit_timer:
      self._re_edit_timer = delayed(self.edit, self.o)
      
  def edit(self, o, ignored_attrs = None):
    if self._re_edit_timer:
      self._re_edit_timer.stop()
      self._re_edit_timer = None
    #if self.o is o: return
    self.can_expand = 0
    
    AttributePane.edit(self, o, ignored_attrs)
    #if isinstance(self.master, EditorPane):
      #if self.can_expand: self.master.box.set_child_packing(self, 1, 1, 0, qt.PackType.START)
      #else:               self.master.box.set_child_packing(self, 0, 1, 0, qt.PackType.START)
      #self.master.box.show_all()
      #min_size, natural_size = self.master.box.get_preferred_size()
      #if self.master.direction == "h": self.master.scroll2.set_min_content_height(min(natural_size.height, 800, _screen.height() - 128))
      #else:                            self.master.scroll2.set_min_content_height(min(natural_size.height, 500, _screen.height() - 128))
    #else: self.show_all()
    
  def _delete_all_fields(self):
    layout = self.layout()
    for i in reversed(range(layout.count())): 
      layout.itemAt(i).widget().setParent(None)
      
  def _new_field(self, attribute, field_class, unit, i):
    label = qtwidgets.QLabel(attribute.label)
    field = field_class(self.gui, self, self.o, attribute, self.undo_stack)
    #label.set_alignment(0.0, 0.0)
    #label.set_margin_top(5)
    self.grid_layout.addWidget(label, i, 0, 1, 1)
    self.grid_layout.addWidget(field, i, 1, 1, 1)
    #field.set_hexpand(1)
    if unit:
      unit_label = qtwidgets.QLabel(unit)
      #unit_label.set_alignment(0.0, 0.5)
      self.grid_layout.addWidget(unit_label, i, 2, 1, 1)
    #if field.y_flags & qt.AttachOptions.EXPAND:
      #field.set_vexpand(1)
    #  self.can_expand = 1
    return field


class OverlayedIconLayout(qtwidgets.QLayout):
  def __init__(self, icon_pane):
    qtwidgets.QLayout.__init__(self)
    self.icon_pane = icon_pane
    self._items = []
    
  def count(self): return len(self._items)
  def addItem(self, item): self._items.append(item)
  def itemAt(self, i):
    if 0 <= i < len(self._items): return self._items[i]
  def takeAt(self, i):
    if 0 <= i < len(self._items):
      item = self._items[i]
      del self._items[i]
      return item
    
  def setGeometry(self, rect):
    x = y = 10
    for item in self._items:
      size = item.sizeHint()
      item.setGeometry(qtcore.QRect(rect.left() + x, rect.top() + y, size.width(), size.height()))
      x += 30
      y += 15
      
  def sizeHint(self): return self.minimumSize()
  def minimumSize(self):
    x = y = 20
    max_width = max_height = 0
    for item in self._items:
      size = item.sizeHint()
      if x + size.width()  > max_width : max_width  = x + size.width ()
      if y + size.height() > max_height: max_height = x + size.height()
      x += 30
      y += 15
    return qtcore.QSize(max_width, max_height)
    

class QtIconPane(IconPane, qtwidgets.QWidget):
  def __init__(self, gui, master):
    qtwidgets.QWidget.__init__(self)
    super().__init__(gui, master)
    
    self.label = qtwidgets.QLabel()
    self.label.setWordWrap(True)
    self.label.setAlignment(qtcore.Qt.AlignTop | qtcore.Qt.AlignJustify)
    self.label.setTextFormat(qtcore.Qt.RichText)
    self.label.setTextInteractionFlags(qtcore.Qt.TextSelectableByMouse | qtcore.Qt.TextSelectableByKeyboard | qtcore.Qt.LinksAccessibleByMouse)
    self.destroyed.connect(self._destroyed)
    self.images = []
    
    layout = qtwidgets.QHBoxLayout()
    self.icons_layout = OverlayedIconLayout(self)
    layout.addLayout(self.icons_layout)
    layout.addWidget(self.label, 1)
    self.setLayout(layout)
    
  def set_nb_images(self, nb):
    if   nb < len(self.images):
      for image in self.images[nb:]: image.setParent(None)
      self.images = self.images[:nb]
      
    elif nb > len(self.images):
      for i in range(nb - len(self.images)):
        image = qtwidgets.QLabel()
        self.images.append(image)
        self.icons_layout.addWidget(image)
        
  def _set_icon_filename_label_details(self, icon_filename, label, details):
    if details: self.label.setText(("<b>%s</b><br/><br/>%s\n" % (escape(label), escape(details))).replace("\n", "<br/>"))
    else:       self.label.setText(("<b>%s</b>"               %  escape(label)).replace("\n", "<br/>"))
    
    self.label.setMinimumSize(self.label.sizeHint())
    
    if   not icon_filename: self.set_nb_images(0)
    
    elif isinstance(icon_filename, str):
      self.set_nb_images(1)
      self.images[0].setPixmap(load_big_icon(icon_filename))
      
    else:
      self.set_nb_images(len(icon_filename))
      icon_filename.sort()
      icon_filename.reverse()
      i = 0
      for filename in icon_filename:
        self.images[i].setPixmap(load_big_icon(filename))
        i += 1

def _load_qt_icon(name):
  icon = qtgui.QIcon.fromTheme(name)
  if icon.isNull():
    return qtgui.QIcon(os.path.join(editobj3._ICON_DIR, "%s.svg" % name))
  return icon
  
        
class QtChildhoodPane(ChildhoodPane, qtwidgets.QWidget):
  def __init__(self, gui, master, undo_stack = None, restrict_to_attribute = None):
    qtwidgets.QWidget.__init__(self)
    super().__init__(gui, master, undo_stack, restrict_to_attribute)
    self.setAutoFillBackground(True)
    self.button_move_up   = qtwidgets.QToolButton()
    self.button_add       = qtwidgets.QToolButton()
    self.button_remove    = qtwidgets.QToolButton()
    self.button_move_down = qtwidgets.QToolButton()
    self.button_move_up  .setIcon(_load_qt_icon("up"    ))
    self.button_add      .setIcon(_load_qt_icon("add"   ))
    self.button_remove   .setIcon(_load_qt_icon("remove"))
    self.button_move_down.setIcon(_load_qt_icon("down"  ))
    self.button_move_up  .setAutoRaise(True)
    self.button_add      .setAutoRaise(True)
    self.button_remove   .setAutoRaise(True)
    self.button_move_down.setAutoRaise(True)
    self.button_move_up  .clicked.connect(self.on_move_up)
    self.button_add      .clicked.connect(self.on_add)
    self.button_remove   .clicked.connect(self.on_remove)
    self.button_move_down.clicked.connect(self.on_move_down)
    #icon_size = qtcore.QSize(editobj3.editor_qt.SMALL_ICON_SIZE, editobj3.editor_qt.SMALL_ICON_SIZE)
    #self.button_move_up  .setSizeHint(icon_size)
    #self.button_add      .setIconSize(icon_size)
    #self.button_remove   .setIconSize(icon_size)
    #self.button_move_down.setIconSize(icon_size)
    
    layout = qtwidgets.QVBoxLayout()
    layout.setContentsMargins(0, 0, 0, 0)
    layout.addWidget(self.button_move_up)
    layout.addWidget(self.button_add)
    layout.addWidget(self.button_remove)
    layout.addWidget(self.button_move_down)
    self.setLayout(layout)
    
  def set_button_visibilities(self, visible1, visible2, visible3, visible4):
    if visible1 or visible2 or visible3 or visible4:
      self.button_move_up  .setVisible(bool(visible1))
      self.button_add      .setVisible(bool(visible2))
      self.button_remove   .setVisible(bool(visible3))
      self.button_move_down.setVisible(bool(visible4))
      self.setVisible(True)
    else:
      self.setVisible(False)


class DynamicTree(qtcore.QAbstractItemModel):
  def set_root_node(self, root_node):
    self.root_node = root_node
    self.children  = [root_node]
    self.o_has_any_children = True

  def expanded(self): return self.children
    
  def parent(self, index):
    if not index.isValid(): return qtcore.QModelIndex()
    node = index.internalPointer()
    #if node.parent is self.root_node: return qtcore.QModelIndex()
    if node is self: return qtcore.QModelIndex()
    if node is self.root_node: return self.createIndex(0, 0, self)
    return self.createIndex(node.parent.index, 0, node.parent)
  
  def hasChildren(self, index):
    if not index.isValid(): return True
    return index.internalPointer().o_has_any_children
  
  def rowCount(self, index):
    if not index.isValid(): return 1
    return len(index.internalPointer().expanded())
  
  def columnCount(self, index): return 1
  
  def index(self, row, column, parent_index):
    if not self.hasIndex(row, column, parent_index): return qtcore.QModelIndex()
    if parent_index.isValid(): parent = parent_index.internalPointer()
    else:                      parent = self
    return self.createIndex(row, column, parent.children[row])
  
  def data(self, index, role):
    if index.isValid():
      if   role == qtcore.Qt.DisplayRole:    return str(index.internalPointer())
      elif role == qtcore.Qt.DecorationRole: return index.internalPointer().get_icon()
    return qtcore.QVariant()

class DynamicFlatList(DynamicTree):
  def set_root_node(self, root_node): self.root_node = root_node
    
  def rowCount(self, index):
    if not index.isValid(): return len(self.root_node.expanded())
    return len(index.internalPointer().expanded())
  
  def parent(self, index):
    if not index.isValid(): return qtcore.QModelIndex()
    node = index.internalPointer()
    if node is self.root_node: return qtcore.QModelIndex()
    return self.createIndex(node.parent.index, 0, node.parent)
  
  def index(self, row, column, parent_index):
    if not self.hasIndex(row, column, parent_index): return qtcore.QModelIndex()
    if parent_index.isValid(): parent = parent_index.internalPointer()
    else:                      parent = self.root_node
    return self.createIndex(row, column, parent.children[row])

  
class DynamicNode(object):
  def __init__(self, parent):
    self.children         = []
    self.children_created = False
    self.is_expandable    = False
    self.index            = 0
    if isinstance(parent, DynamicNode):
      self.parent         = parent
      self.tree           = parent.tree
    else:
      self.parent         = None
      self.tree           = parent
      self.update()
      self.update_children()
      
  def has_children(self): raise NotImplementedError
  def create_children(self, old_children = None): return []
  
  def qt_index(self): return self.tree.createIndex(self.index, 0, self)
  
  def expand(self): self.tree.view.expand(self.qt_index())
  
  def expanded(self):
    if not self.children_created:
      self.children_created = True
      self.update_children()
    return self.children
  
  def collapsed(self):
    if self.children_created:
      self.children_created = False
      for child in self.children[::-1]: child.destroy()
      self.is_expandable = False
      self.update_children()
      
  def update(self): pass
  
  def update_children(self):
    if self.children_created:
      old_children = self.children[:]
      new_children = self.create_children(old_children)
      old_children_set = set(old_children)
      new_children_set = set(new_children)
      
      for child in old_children[::-1]:
        if not child in new_children_set: child.destroy()
        
      kept_children = [child for child in new_children if child in old_children_set]
      
      for i in range(len(self.children)): self.children[i].index     = i
      for i in range(len(kept_children)): kept_children[i].new_index = i
      
      qt_index = self.qt_index()
      if kept_children:
        for child in kept_children:
          child_index = self.children.index(child)
          if child.new_index == child_index : continue # Already at the right position
          
          self.tree.beginMoveRows(qt_index, child_index, child_index, qt_index, child.new_index)
          child.index = child.new_index
          del self.children[child_index]
          self.children.insert(child.new_index, child)
          self.tree.endMoveRows()
          
        for child in kept_children: del child.new_index
        
      for i in range(len(new_children)):
        new_children[i].index = i
        if not new_children[i] in old_children_set:
          self.tree.beginInsertRows(qt_index, i, i)
          self.children.insert(i, new_children[i])
          self.tree.endInsertRows()
          
      for child in new_children:
        if not child in old_children_set:
          child.update()
          child.update_children()
          
      self.children = new_children
      
    else:
      if self.flat_list: pass
      else:              self.is_expandable = bool(self.has_children())
      
  def destroy(self):
    for child in self.children[::-1]: child.destroy()
    if self.parent and (self.tree.view.tree is self.tree):
      qt_parent_index = self.parent.qt_index()
      self.tree.beginRemoveRows(qt_parent_index, self.index, self.index)
      self.parent.children.remove(self)
      self.tree.endRemoveRows()
    

class Qt_HierarchyNode(HierarchyNode, DynamicNode):
  def update(self):
    icon_filename = self.descr.icon_filename_for(self.o)
    #if isinstance(icon_filename, str): icon = load_small_icon(icon_filename)
    #else:                              icon = None
    index = self.qt_index()
    self.tree.dataChanged.emit(index, index)
    
  def get_icon(self):
    icon_filename = self.descr.icon_filename_for(self.o)
    if isinstance(icon_filename, str): return load_small_icon(icon_filename)
    
  def __str__(self):
    if self.attribute and not self.restrict_to_attribute: attr_label = self.attribute.label
    else:                                                 attr_label = ""
    if attr_label: return '''%s : %s''' % (attr_label, self.descr.label_for(self.o))
    return self.descr.label_for(self.o)
  
# class HierarchyScrollBar(qtwidgets.QScrollBar):
#   def __init__(self, hierarchy_pane):
#     qtwidgets.QScrollBar.__init__(self, qtcore.Qt.Vertical)
#     self.hierarchy_pane = hierarchy_pane

#   def hideEvent(self, event):
#     qtwidgets.QScrollBar.hideEvent(self, event)
#     print("cachée")
#     #self..overlay_offset_x = 3

#   def showEvent(self, event):
#     qtwidgets.QScrollBar.showEvent(self, event)
#     print("visible")
#     #self..overlay_offset_x = 3 + self.

class QtHierarchyPane(HierarchyPane, qtwidgets.QTreeView):
  Node = Qt_HierarchyNode
  def __init__(self, gui, master, edit_child, undo_stack = None, restrict_to_attribute = None, flat_list = 0):
    qtwidgets.QTreeView.__init__(self)
    
    self.setSelectionMode(qtwidgets.QAbstractItemView.ExtendedSelection)
    self.setHeaderHidden(True)
    self.setIconSize(qtcore.QSize(editobj3.editor_qt.SMALL_ICON_SIZE, editobj3.editor_qt.SMALL_ICON_SIZE))
    
    super().__init__(gui, master, edit_child, undo_stack, restrict_to_attribute, flat_list)
    self.destroyed.connect(self._destroyed)
    self.selected_parent     = None
    self.selected_attribute  = None
    self.selected_node       = None
    self.selected            = None
    self.last_button         = 0
    
    self.collapsed.connect(self.on_collapsed)
    self.pressed.connect(self.on_button_pressed)
    self.doubleClicked.connect(self.on_double_click)
    
  def edit(self, *args):
    if len(args) >= 3: return qtwidgets.QTreeView.edit(self, *args)
    o = args[0]
    if o is self.o: return

    if self.flat_list: self.tree = DynamicFlatList(); self.setRootIsDecorated(False)
    else:              self.tree = DynamicTree()
    self.tree.view = self
    
    
    HierarchyPane.edit(self, o)
    self.tree.set_root_node(self.root_node)
    self.root_node.expanded()
    self.setModel(self.tree)
    self.selectionModel().selectionChanged.connect(self.on_selection_changed)
    if self.flat_list:
      self.childhood_pane.edit(self.childhood_pane_parent, self.childhood_pane_attribute, self.o)
    else:
      self.setExpanded(self.root_node.qt_index(), True)
      self.selectionModel().select(self.root_node.qt_index(), qtcore.QItemSelectionModel.ClearAndSelect)
      
  def expand_tree_at_level(self, level): self.expandToDepth(level)
  
  def select_node(self, node):
    self.selectionModel().clearSelection()
    if   node:                                   self.selectionModel().select(node.qt_index(), qtcore.QItemSelectionModel.Select)
    elif self.flat_list and self.childhood_pane: self.childhood_pane.edit(self.childhood_pane_parent, self.childhood_pane_attribute, self.o)
    
  def on_collapsed(self, index):
    node = index.internalPointer()
    node.collapsed()
    
  def on_double_click(self, index):
    actions = self.get_actions(self.selected_parent, self.selected_attribute, self.selected)
    for action in actions:
      if action.default:
        self._action_activated(action, self.selected_parent, self.selected_attribute, self.selected)
        break
      
  def on_button_pressed(self, index):
    if qtgui.QGuiApplication.mouseButtons() == qtcore.Qt.RightButton:
      actions = self.get_actions(self.selected_parent, self.selected_attribute, self.selected)
      self.show_action_menu(actions)
      
  def on_selection_changed(self, selected, deselected):
    indexes = self.selectionModel().selectedIndexes()
    if len(indexes) == 0:
      self.edit_child(None)
      if self.childhood_pane: self.childhood_pane.edit(self.childhood_pane_parent, self.childhood_pane_attribute, None)
      self.selected_node = None
      return
    
    if len(indexes) == 1:
      node = self.selected_node = indexes[0].internalPointer()
      if node.parent: self.selected_parent = node.parent.o
      else:           self.selected_parent = None
      self.selected           = node.o
      self.selected_attribute = node.attribute
      
      self.edit_child(self.selected, self.get_ignored_attrs([node]))
      if self.childhood_pane: self.childhood_pane.edit(self.selected_parent, self.selected_attribute, self.selected)
      
    else:
      nodes = self.selected_node = [index.internalPointer() for index in indexes]
      self.selected_parent    = introsp.ObjectPack([(node.parent and node.parent.o) for node in nodes])
      self.selected           = introsp.ObjectPack([node.o for node in nodes])
      self.selected_attribute = [node.attribute for node in nodes]
      
      self.edit_child(self.selected, self.get_ignored_attrs(nodes))
      if self.childhood_pane: self.childhood_pane.edit(self.selected_parent, self.selected_attribute, self.selected)
      
  def on_action_activated(self, action):
    if self.selected:
      self._action_activated(action, self.selected_parent, self.selected_attribute, self.selected)
    else:
      self._action_activated(action, None, self.restrict_to_attribute, self.o)
      
  def show_action_menu(self, actions):
    self._popup_menu = qtwidgets.QMenu() # The menu must be kept in memory, else it does not appear.
    for action in actions:
      menu_item = qtwidgets.QAction(action.label, self._popup_menu)
      menu_item.triggered.connect(lambda checked, action = action: self.on_action_activated(action))
      if action.default:
        font = qtgui.QFont(menu_item.font())
        font.setBold(True)
        menu_item.setFont(font)
      self._popup_menu.addAction(menu_item)
    self._popup_menu.popup(qtgui.QCursor.pos())

  
SMALL_ICON_SIZE = 32

_small_icons = {}
_big_icons   = {}

def load_big_icon(filename):
  pixmap = _big_icons.get(filename)
  if not pixmap: pixmap = _big_icons[filename] = qtgui.QPixmap(filename)
  return pixmap

def load_small_icon(filename):
  icon = _small_icons.get(filename)
  if not icon: icon = _small_icons[filename] = qtgui.QIcon(filename)
  return icon



# -*- coding: utf-8 -*-
# Editobj3
# Copyright (C) 2007-2014 Jean-Baptiste LAMY

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, os, os.path, time, _thread, editobj3, editobj3.introsp as introsp, editobj3.observe as observe, editobj3.field as field
import editobj3.undoredo as undoredo

class Download(object):
  def __init__(self, remote_url, local_filename):
    self.remote_url     = remote_url
    self.local_filename = local_filename
    self.progress       = 0.0
    self.speed          = 0
    self.completed      = 0
    
  def __str__(self): return os.path.basename(self.remote_url) + " (" + str(int(100 * self.progress)) + "%)"

  def start(self):
    self.progress = 0.0
    self.speed    = 10
    
  def stop(self): pass


class DownloadManager(object):
  def __init__(self):
    self.downloads = []
    
  def __str__(self): return "Download manager"

  def add_download(self, download):
    self.insert_download(len(self.downloads) + 1, download)

  def insert_download(self, index, download):
    self.downloads.insert(index, download)
    download.start()

  def remove_download(self, download):
    download.stop()
    self.downloads.remove(download)

icon_filename = os.path.join(os.path.dirname(sys.argv[0]), "./jiba.png")

dm = DownloadManager()
dm.add_download(Download("http://python.org", "./index_python.html"))
dm.add_download(Download("http://soyaproject.org/soya", "./index_soya.html"))
dm.add_download(Download("http://soyaproject.org/slune", "./index_slune.html"))
dm.add_download(Download("http://soyaproject.org/balazar_brother", "./index_balazar_brother.html"))

def downloader():
  for download in dm.downloads[:]:
    if download.progress < 1.0:
      download.progress = download.progress + 0.001 * download.speed
      if download.progress >= 1.0:
        download.progress = 1.0
        download.completed = 1
  observe.scan()
  return 1

def run_downloader():
  import time
  time.sleep(1.0) # Wait for the app to start up
  while 1:
    downloader()
    time.sleep(0.2)
  
descr = introsp.description(Download)
descr.def_attr("progress", field.ProgressBarField)
descr.def_attr("speed", field.RangeField(0, 100), unit = "Ko/s")
descr.def_attr("completed", field.BoolField)
descr.set_icon_filename(os.path.join(os.path.dirname(sys.argv[0]), "./file.png"))

class NewDowloadFormConstructor(introsp.FormConstructor):
  def __init__(self):
    self.remote_url     = "http://"
    self.local_filename = ""

# You may want to customize the creation of the new instance -- by default, EditObj 3
# uses the form's __dict__ as named args.

#  def create_new_from_form(self, klass, parent):
#    return klass(self.remote_url, self.local_filename)

descr.set_constructor(NewDowloadFormConstructor)

descr = introsp.description(DownloadManager)
descr.set_details("To add a new download, click on the '+' button on the right.\nUse the '-' button to remove and cancel a download.")

#download_constructor = introsp.Constructor(Download, lambda download_manager: Download("http://", ""))
descr.def_attr("downloads", addable_values = [introsp.NewInstanceOf(Download)])

descr.add_action(introsp.Action("Undo", lambda undo_stack, editor, o: undo_stack.undo()))
descr.add_action(introsp.Action("Redo", lambda undo_stack, editor, o: undo_stack.redo()))

if len(sys.argv) > 1: editobj3.GUI = sys.argv[-1]

w = editobj3.edit(dm, direction = "v", on_close = sys.exit)


if editobj3.GUI == "Gtk":
  import gi
  from gi.repository import GObject as gobject
  gobject.timeout_add(200, downloader) # Gtk does not like being used along with Python thread module
else:
  import PyQt5.QtCore as qtcore
  timer = qtcore.QTimer()
  timer.timeout.connect(downloader)
  timer.setInterval(200)
  timer.start()
  #  import _thread; _thread.start_new_thread(run_downloader, ())
  pass
  
w.main()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# Editobj3
# Copyright (C) 2007-2014 Jean-Baptiste LAMY

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os, os.path, sys, glob

HERE = os.path.relpath(os.path.dirname(os.path.abspath(__file__)))

if len(sys.argv) <= 1: sys.argv.append("install")


import setuptools

#data_files = [os.path.join("js", file) for file in os.listdir("js") if (file != "CVS") and (file != ".hg") and (file != ".svn") and (file != ".arch-ids")] \
#           + [os.path.join("icons", file) for file in os.listdir("icons") if (file != "CVS") and (file != ".hg") and (file != ".svn") and (file != ".arch-ids")]

setuptools.setup(
  name         = "Editobj3",
  version      = "0.2",
  license      = "LGPLv3+",
  description  = "An automatic dialog box generator for Python objects, supporting multiple graphical backends: Qt, GTK and HTML (single-user or with multiple users).",
  long_description = open(os.path.join(HERE, "README.rst")).read(),
  
  author       = "Lamy Jean-Baptiste (Jiba)",
  author_email = "jibalamy@free.fr",
  url          = "https://bitbucket.org/jibalamy/editobj3",
  
  classifiers  = [
    "Development Status :: 4 - Beta",
    "Intended Audience :: Developers",
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
    "Operating System :: OS Independent",
    "Environment :: Web Environment",
    "Environment :: X11 Applications :: Qt",
    "Environment :: X11 Applications :: GTK",
    "Topic :: Internet :: WWW/HTTP :: HTTP Servers",
    "Topic :: Scientific/Engineering :: Human Machine Interfaces",
    "Topic :: Software Development :: User Interfaces",
    "Topic :: Software Development :: Libraries :: Python Modules",
    ],
  
  package_dir  = {"editobj3" : HERE},
  packages     = ["editobj3"],
  
  package_data = {"editobj3" : ["icons/*", "js/*"]}
  )



# -*- coding: utf-8 -*-
# Editobj3
# Copyright (C) 2007-2014 Jean-Baptiste LAMY

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
editobj3.undoredo: Multiple undo/redo framework
-----------------------------------------------

This module contains a multiple undo/redo framework. It is used by Editobj3 dialog boxes,
and it automatically call :func:`editobj3.observe.scan()` when doing or undoing an operation.

.. data:: stack
   :annotation: the default undo/redo stack.
"""

__all__ = ["Stack", "UndoableOperation"]

import editobj3.observe as observe

class Stack(object):
  """An undo/redo stack.

:param limit: the maximum number of undo/redo, defaults to 20."""
  def __init__(self, limit = 20):
    self.limit     = limit
    self.undoables = []
    self.redoables = []
      
  def can_undo(self):
    """Returns True if it is possible to undo an operation."""
    if self.undoables: return self.undoables[-1]
    return False
  
  def can_redo(self):
    """Returns True if it is possible to redo an operation."""
    if self.redoables: return self.redoables[-1]
    return False
  
  def do_operation(self, operation):
    """Does the operation. Can be overriden, e.g. to save data after the changes performed by the operation.

:param operation: the operation.
:type operation: :class:`Operation`"""
    operation.do_func()
    
  def undo(self):
    """Undoes the last operation available."""
    if not self.undoables: raise ValueError("No operation to undo!")
    undo = self.undoables.pop()
    opposite = undo.opposite()
    
  def redo(self):
    """Redoes the last operation available."""
    if not self.redoables: raise ValueError("No operation to redo!")
    redo = self.redoables.pop()
    opposite = redo.opposite()
    
  def clear(self):
    """Clears all undo/redo operations."""
    self.undoables = []
    self.redoables = []

  def merge_last_operations(self, name = "", nb = 2):
    """Merges the NB last operations. They will now be undone / redone as a single operation, with the given NAME."""
    if not name: name = ", ".join(undoable.name for undoable in self.undoables[-nb:])
    doers   = [undoable.do_func   for undoable in self.undoables[-nb:]]
    undoers = [undoable.undo_func for undoable in self.undoables[-nb:]]
    def do_it():
      for doer in doers: doer()
    def undo_it():
      for undoer in reversed(undoers): undoer()
      
    del self.undoables[-nb + 1:]
    self.undoables[-1].do_func   = do_it
    self.undoables[-1].undo_func = undo_it
    self.undoables[-1].name      = name
    return self.undoables[-1]
  
  def __repr__(self):
    return "<%s, undoables:\n%s\n  redoables:\n%s\n>" % (
      self.__class__.__name__,
      "\n".join(["    %s" % repr(i) for i in self.undoables]),
      "\n".join(["    %s" % repr(i) for i in self.redoables]),
      )
    
stack = Stack()


class _Operation(object):
  def __init__(self, do_func, undo_func, name = "", stack_ = stack):
    self.do_func   = do_func
    self.undo_func = undo_func
    self.name      = name
    self.stack     = stack_ or stack
    stack.do_operation(self)
    
  def __repr__(self):
    return "<%s '%s' do_func='%s' undo_func='%s'>" % (self.__class__.__name__, self.name, self.do_func, self.undo_func)
    
class UndoableOperation(_Operation):
  """UndoableOperation(do_func, undo_func, name = "", stack = undoredo.stack)

An operation that can be undone.

:param do_func: a callable that do the operation when called.
:param undo_func: a callable that undo the operation when called.
:param name: the name of the operation.
:param stack: the undo/redo stack to add the operation to, defaults to undoredo.stack.
"""
  def __init__(self, do_func, undo_func, name = "", stack = None):
    _Operation.__init__(self, do_func, undo_func, name, stack)
    stack.undoables.append(self)
    if len(self.stack.undoables) > self.stack.limit: del self.stack.undoables[0]
    observe.scan()
    
  def opposite(self):
    return _RedoableOperation(self.undo_func, self.do_func, self.name, self.stack)

  def coalesce_with(self, previous_undoable_operation):
    self.undo_func = previous_undoable_operation.undo_func
    previous_undoable_operation.stack.undoables.remove(previous_undoable_operation)
    
class _RedoableOperation(_Operation):
  def __init__(self, do_func, undo_func, name = "", stack = stack):
    _Operation.__init__(self, do_func, undo_func, name, stack)
    stack.redoables.append(self)
    observe.scan()
    
  def opposite(self):
    return UndoableOperation(self.undo_func, self.do_func, self.name, self.stack)


